"""
Component library for SBYB UI Generator.

This module provides a library of reusable UI components
for building dashboards and forms.
"""

from typing import Any, Dict, List, Optional, Union, Tuple
import os
import json
import shutil
import tempfile

from sbyb.core.base import SBYBComponent
from sbyb.core.exceptions import UIGenerationError


class ComponentLibrary(SBYBComponent):
    """
    Component library for UI Generator.
    
    This component provides a library of reusable UI components
    for building dashboards and forms.
    """
    
    def __init__(self, config: Optional[Dict[str, Any]] = None):
        """
        Initialize the component library.
        
        Args:
            config: Configuration dictionary for the library.
        """
        super().__init__(config)
        self.components = {}
        self.categories = {}
        self._load_default_components()
    
    def _load_default_components(self) -> None:
        """
        Load default components.
        """
        # Data visualization components
        self.add_category("data_visualization", "Data Visualization")
        
        self.register_component(
            "bar_chart",
            "Bar Chart",
            "data_visualization",
            {
                "description": "A bar chart component for visualizing categorical data.",
                "properties": {
                    "data_source": {"type": "string", "required": True},
                    "x_axis": {"type": "string", "required": True},
                    "y_axis": {"type": "string", "required": True},
                    "title": {"type": "string", "required": False},
                    "color": {"type": "string", "required": False},
                    "orientation": {"type": "string", "enum": ["vertical", "horizontal"], "default": "vertical"}
                },
                "frameworks": {
                    "streamlit": "st.bar_chart",
                    "dash": "px.bar",
                    "plotly": "go.Bar",
                    "matplotlib": "plt.bar"
                }
            }
        )
        
        self.register_component(
            "line_chart",
            "Line Chart",
            "data_visualization",
            {
                "description": "A line chart component for visualizing trends over time.",
                "properties": {
                    "data_source": {"type": "string", "required": True},
                    "x_axis": {"type": "string", "required": True},
                    "y_axis": {"type": "string", "required": True},
                    "title": {"type": "string", "required": False},
                    "color": {"type": "string", "required": False},
                    "markers": {"type": "boolean", "default": False}
                },
                "frameworks": {
                    "streamlit": "st.line_chart",
                    "dash": "px.line",
                    "plotly": "go.Scatter",
                    "matplotlib": "plt.plot"
                }
            }
        )
        
        self.register_component(
            "scatter_plot",
            "Scatter Plot",
            "data_visualization",
            {
                "description": "A scatter plot component for visualizing relationships between variables.",
                "properties": {
                    "data_source": {"type": "string", "required": True},
                    "x_axis": {"type": "string", "required": True},
                    "y_axis": {"type": "string", "required": True},
                    "title": {"type": "string", "required": False},
                    "color": {"type": "string", "required": False},
                    "size": {"type": "string", "required": False},
                    "hover_data": {"type": "array", "items": {"type": "string"}, "required": False}
                },
                "frameworks": {
                    "streamlit": "st.scatter_chart",
                    "dash": "px.scatter",
                    "plotly": "go.Scatter",
                    "matplotlib": "plt.scatter"
                }
            }
        )
        
        self.register_component(
            "pie_chart",
            "Pie Chart",
            "data_visualization",
            {
                "description": "A pie chart component for visualizing proportions.",
                "properties": {
                    "data_source": {"type": "string", "required": True},
                    "names": {"type": "string", "required": True},
                    "values": {"type": "string", "required": True},
                    "title": {"type": "string", "required": False},
                    "hole": {"type": "number", "default": 0}
                },
                "frameworks": {
                    "streamlit": "st.pyplot",
                    "dash": "px.pie",
                    "plotly": "go.Pie",
                    "matplotlib": "plt.pie"
                }
            }
        )
        
        self.register_component(
            "histogram",
            "Histogram",
            "data_visualization",
            {
                "description": "A histogram component for visualizing distributions.",
                "properties": {
                    "data_source": {"type": "string", "required": True},
                    "x": {"type": "string", "required": True},
                    "bins": {"type": "integer", "default": 10},
                    "title": {"type": "string", "required": False},
                    "color": {"type": "string", "required": False}
                },
                "frameworks": {
                    "streamlit": "st.pyplot",
                    "dash": "px.histogram",
                    "plotly": "go.Histogram",
                    "matplotlib": "plt.hist"
                }
            }
        )
        
        self.register_component(
            "heatmap",
            "Heatmap",
            "data_visualization",
            {
                "description": "A heatmap component for visualizing correlations or 2D distributions.",
                "properties": {
                    "data_source": {"type": "string", "required": True},
                    "title": {"type": "string", "required": False},
                    "color_scale": {"type": "string", "default": "viridis"},
                    "annotations": {"type": "boolean", "default": True}
                },
                "frameworks": {
                    "streamlit": "st.pyplot",
                    "dash": "px.imshow",
                    "plotly": "go.Heatmap",
                    "matplotlib": "plt.imshow"
                }
            }
        )
        
        self.register_component(
            "box_plot",
            "Box Plot",
            "data_visualization",
            {
                "description": "A box plot component for visualizing distributions and outliers.",
                "properties": {
                    "data_source": {"type": "string", "required": True},
                    "x": {"type": "string", "required": False},
                    "y": {"type": "string", "required": True},
                    "title": {"type": "string", "required": False},
                    "color": {"type": "string", "required": False},
                    "notched": {"type": "boolean", "default": False}
                },
                "frameworks": {
                    "streamlit": "st.pyplot",
                    "dash": "px.box",
                    "plotly": "go.Box",
                    "matplotlib": "plt.boxplot"
                }
            }
        )
        
        # Input components
        self.add_category("input", "Input Components")
        
        self.register_component(
            "text_input",
            "Text Input",
            "input",
            {
                "description": "A text input component for collecting string data.",
                "properties": {
                    "label": {"type": "string", "required": True},
                    "default": {"type": "string", "required": False},
                    "placeholder": {"type": "string", "required": False},
                    "help_text": {"type": "string", "required": False},
                    "required": {"type": "boolean", "default": False},
                    "max_length": {"type": "integer", "required": False}
                },
                "frameworks": {
                    "streamlit": "st.text_input",
                    "dash": "dcc.Input",
                    "html": "input[type='text']",
                    "react": "TextInput",
                    "vue": "TextInput"
                }
            }
        )
        
        self.register_component(
            "number_input",
            "Number Input",
            "input",
            {
                "description": "A number input component for collecting numeric data.",
                "properties": {
                    "label": {"type": "string", "required": True},
                    "default": {"type": "number", "required": False},
                    "min": {"type": "number", "required": False},
                    "max": {"type": "number", "required": False},
                    "step": {"type": "number", "default": 1},
                    "help_text": {"type": "string", "required": False},
                    "required": {"type": "boolean", "default": False}
                },
                "frameworks": {
                    "streamlit": "st.number_input",
                    "dash": "dcc.Input",
                    "html": "input[type='number']",
                    "react": "NumberInput",
                    "vue": "NumberInput"
                }
            }
        )
        
        self.register_component(
            "slider",
            "Slider",
            "input",
            {
                "description": "A slider component for selecting a value from a range.",
                "properties": {
                    "label": {"type": "string", "required": True},
                    "min": {"type": "number", "default": 0},
                    "max": {"type": "number", "default": 100},
                    "default": {"type": "number", "required": False},
                    "step": {"type": "number", "default": 1},
                    "help_text": {"type": "string", "required": False}
                },
                "frameworks": {
                    "streamlit": "st.slider",
                    "dash": "dcc.Slider",
                    "html": "input[type='range']",
                    "react": "Slider",
                    "vue": "Slider"
                }
            }
        )
        
        self.register_component(
            "select",
            "Select",
            "input",
            {
                "description": "A select component for choosing from a list of options.",
                "properties": {
                    "label": {"type": "string", "required": True},
                    "options": {"type": "array", "items": {"type": "string"}, "required": True},
                    "default": {"type": "string", "required": False},
                    "help_text": {"type": "string", "required": False},
                    "required": {"type": "boolean", "default": False}
                },
                "frameworks": {
                    "streamlit": "st.selectbox",
                    "dash": "dcc.Dropdown",
                    "html": "select",
                    "react": "Select",
                    "vue": "Select"
                }
            }
        )
        
        self.register_component(
            "multiselect",
            "Multi-Select",
            "input",
            {
                "description": "A multi-select component for choosing multiple options from a list.",
                "properties": {
                    "label": {"type": "string", "required": True},
                    "options": {"type": "array", "items": {"type": "string"}, "required": True},
                    "default": {"type": "array", "items": {"type": "string"}, "required": False},
                    "help_text": {"type": "string", "required": False}
                },
                "frameworks": {
                    "streamlit": "st.multiselect",
                    "dash": "dcc.Dropdown",
                    "html": "select[multiple]",
                    "react": "MultiSelect",
                    "vue": "MultiSelect"
                }
            }
        )
        
        self.register_component(
            "checkbox",
            "Checkbox",
            "input",
            {
                "description": "A checkbox component for boolean input.",
                "properties": {
                    "label": {"type": "string", "required": True},
                    "default": {"type": "boolean", "default": False},
                    "help_text": {"type": "string", "required": False}
                },
                "frameworks": {
                    "streamlit": "st.checkbox",
                    "dash": "dcc.Checklist",
                    "html": "input[type='checkbox']",
                    "react": "Checkbox",
                    "vue": "Checkbox"
                }
            }
        )
        
        self.register_component(
            "radio",
            "Radio Buttons",
            "input",
            {
                "description": "A radio button component for selecting one option from a list.",
                "properties": {
                    "label": {"type": "string", "required": True},
                    "options": {"type": "array", "items": {"type": "string"}, "required": True},
                    "default": {"type": "string", "required": False},
                    "help_text": {"type": "string", "required": False},
                    "required": {"type": "boolean", "default": False}
                },
                "frameworks": {
                    "streamlit": "st.radio",
                    "dash": "dcc.RadioItems",
                    "html": "input[type='radio']",
                    "react": "RadioGroup",
                    "vue": "RadioGroup"
                }
            }
        )
        
        self.register_component(
            "date_input",
            "Date Input",
            "input",
            {
                "description": "A date input component for selecting dates.",
                "properties": {
                    "label": {"type": "string", "required": True},
                    "default": {"type": "string", "required": False},
                    "min": {"type": "string", "required": False},
                    "max": {"type": "string", "required": False},
                    "help_text": {"type": "string", "required": False},
                    "required": {"type": "boolean", "default": False}
                },
                "frameworks": {
                    "streamlit": "st.date_input",
                    "dash": "dcc.DatePickerSingle",
                    "html": "input[type='date']",
                    "react": "DatePicker",
                    "vue": "DatePicker"
                }
            }
        )
        
        self.register_component(
            "file_upload",
            "File Upload",
            "input",
            {
                "description": "A file upload component for uploading files.",
                "properties": {
                    "label": {"type": "string", "required": True},
                    "accept": {"type": "string", "required": False},
                    "multiple": {"type": "boolean", "default": False},
                    "help_text": {"type": "string", "required": False}
                },
                "frameworks": {
                    "streamlit": "st.file_uploader",
                    "dash": "dcc.Upload",
                    "html": "input[type='file']",
                    "react": "FileUpload",
                    "vue": "FileUpload"
                }
            }
        )
        
        # Layout components
        self.add_category("layout", "Layout Components")
        
        self.register_component(
            "tabs",
            "Tabs",
            "layout",
            {
                "description": "A tabs component for organizing content into tabs.",
                "properties": {
                    "tabs": {"type": "array", "items": {"type": "object"}, "required": True}
                },
                "frameworks": {
                    "streamlit": "st.tabs",
                    "dash": "dcc.Tabs",
                    "html": "div.tabs",
                    "react": "Tabs",
                    "vue": "Tabs"
                }
            }
        )
        
        self.register_component(
            "columns",
            "Columns",
            "layout",
            {
                "description": "A columns component for creating multi-column layouts.",
                "properties": {
                    "columns": {"type": "array", "items": {"type": "object"}, "required": True},
                    "widths": {"type": "array", "items": {"type": "number"}, "required": False}
                },
                "frameworks": {
                    "streamlit": "st.columns",
                    "dash": "html.Div",
                    "html": "div.columns",
                    "react": "Grid",
                    "vue": "Grid"
                }
            }
        )
        
        self.register_component(
            "expander",
            "Expander",
            "layout",
            {
                "description": "An expander component for collapsible content.",
                "properties": {
                    "label": {"type": "string", "required": True},
                    "expanded": {"type": "boolean", "default": False}
                },
                "frameworks": {
                    "streamlit": "st.expander",
                    "dash": "html.Details",
                    "html": "details",
                    "react": "Accordion",
                    "vue": "Accordion"
                }
            }
        )
        
        self.register_component(
            "container",
            "Container",
            "layout",
            {
                "description": "A container component for grouping content.",
                "properties": {
                    "border": {"type": "boolean", "default": False},
                    "padding": {"type": "string", "default": "1rem"},
                    "background_color": {"type": "string", "required": False}
                },
                "frameworks": {
                    "streamlit": "st.container",
                    "dash": "html.Div",
                    "html": "div.container",
                    "react": "Container",
                    "vue": "Container"
                }
            }
        )
        
        self.register_component(
            "divider",
            "Divider",
            "layout",
            {
                "description": "A divider component for separating content.",
                "properties": {
                    "style": {"type": "string", "enum": ["solid", "dashed", "dotted"], "default": "solid"},
                    "color": {"type": "string", "required": False},
                    "thickness": {"type": "number", "default": 1}
                },
                "frameworks": {
                    "streamlit": "st.divider",
                    "dash": "html.Hr",
                    "html": "hr",
                    "react": "Divider",
                    "vue": "Divider"
                }
            }
        )
        
        # Display components
        self.add_category("display", "Display Components")
        
        self.register_component(
            "text",
            "Text",
            "display",
            {
                "description": "A text component for displaying text.",
                "properties": {
                    "text": {"type": "string", "required": True},
                    "size": {"type": "string", "enum": ["small", "medium", "large"], "default": "medium"},
                    "color": {"type": "string", "required": False},
                    "align": {"type": "string", "enum": ["left", "center", "right"], "default": "left"}
                },
                "frameworks": {
                    "streamlit": "st.text",
                    "dash": "html.P",
                    "html": "p",
                    "react": "Text",
                    "vue": "Text"
                }
            }
        )
        
        self.register_component(
            "header",
            "Header",
            "display",
            {
                "description": "A header component for displaying headings.",
                "properties": {
                    "text": {"type": "string", "required": True},
                    "level": {"type": "integer", "enum": [1, 2, 3, 4, 5, 6], "default": 2},
                    "color": {"type": "string", "required": False},
                    "align": {"type": "string", "enum": ["left", "center", "right"], "default": "left"}
                },
                "frameworks": {
                    "streamlit": "st.header",
                    "dash": "html.H2",
                    "html": "h2",
                    "react": "Heading",
                    "vue": "Heading"
                }
            }
        )
        
        self.register_component(
            "image",
            "Image",
            "display",
            {
                "description": "An image component for displaying images.",
                "properties": {
                    "src": {"type": "string", "required": True},
                    "alt": {"type": "string", "required": False},
                    "width": {"type": "string", "required": False},
                    "height": {"type": "string", "required": False},
                    "caption": {"type": "string", "required": False}
                },
                "frameworks": {
                    "streamlit": "st.image",
                    "dash": "html.Img",
                    "html": "img",
                    "react": "Image",
                    "vue": "Image"
                }
            }
        )
        
        self.register_component(
            "data_table",
            "Data Table",
            "display",
            {
                "description": "A data table component for displaying tabular data.",
                "properties": {
                    "data_source": {"type": "string", "required": True},
                    "columns": {"type": "array", "items": {"type": "string"}, "required": False},
                    "pagination": {"type": "boolean", "default": True},
                    "page_size": {"type": "integer", "default": 10},
                    "sortable": {"type": "boolean", "default": True},
                    "filterable": {"type": "boolean", "default": False}
                },
                "frameworks": {
                    "streamlit": "st.dataframe",
                    "dash": "dash_table.DataTable",
                    "html": "table",
                    "react": "DataTable",
                    "vue": "DataTable"
                }
            }
        )
        
        self.register_component(
            "metric",
            "Metric",
            "display",
            {
                "description": "A metric component for displaying key metrics.",
                "properties": {
                    "label": {"type": "string", "required": True},
                    "value": {"type": "string", "required": True},
                    "delta": {"type": "string", "required": False},
                    "color": {"type": "string", "required": False}
                },
                "frameworks": {
                    "streamlit": "st.metric",
                    "dash": "html.Div",
                    "html": "div.metric",
                    "react": "Metric",
                    "vue": "Metric"
                }
            }
        )
        
        self.register_component(
            "code",
            "Code",
            "display",
            {
                "description": "A code component for displaying code snippets.",
                "properties": {
                    "code": {"type": "string", "required": True},
                    "language": {"type": "string", "default": "python"},
                    "line_numbers": {"type": "boolean", "default": False}
                },
                "frameworks": {
                    "streamlit": "st.code",
                    "dash": "dcc.Markdown",
                    "html": "pre code",
                    "react": "CodeBlock",
                    "vue": "CodeBlock"
                }
            }
        )
        
        self.register_component(
            "markdown",
            "Markdown",
            "display",
            {
                "description": "A markdown component for displaying formatted text.",
                "properties": {
                    "content": {"type": "string", "required": True}
                },
                "frameworks": {
                    "streamlit": "st.markdown",
                    "dash": "dcc.Markdown",
                    "html": "div.markdown",
                    "react": "Markdown",
                    "vue": "Markdown"
                }
            }
        )
        
        # Interactive components
        self.add_category("interactive", "Interactive Components")
        
        self.register_component(
            "button",
            "Button",
            "interactive",
            {
                "description": "A button component for triggering actions.",
                "properties": {
                    "label": {"type": "string", "required": True},
                    "type": {"type": "string", "enum": ["primary", "secondary", "danger"], "default": "primary"},
                    "disabled": {"type": "boolean", "default": False},
                    "icon": {"type": "string", "required": False}
                },
                "frameworks": {
                    "streamlit": "st.button",
                    "dash": "html.Button",
                    "html": "button",
                    "react": "Button",
                    "vue": "Button"
                }
            }
        )
        
        self.register_component(
            "download_button",
            "Download Button",
            "interactive",
            {
                "description": "A download button component for downloading files.",
                "properties": {
                    "label": {"type": "string", "required": True},
                    "data": {"type": "string", "required": True},
                    "file_name": {"type": "string", "required": True},
                    "mime_type": {"type": "string", "required": False}
                },
                "frameworks": {
                    "streamlit": "st.download_button",
                    "dash": "html.A",
                    "html": "a[download]",
                    "react": "DownloadButton",
                    "vue": "DownloadButton"
                }
            }
        )
        
        self.register_component(
            "toggle",
            "Toggle",
            "interactive",
            {
                "description": "A toggle component for switching between two states.",
                "properties": {
                    "label": {"type": "string", "required": True},
                    "default": {"type": "boolean", "default": False},
                    "disabled": {"type": "boolean", "default": False}
                },
                "frameworks": {
                    "streamlit": "st.toggle",
                    "dash": "dcc.Checklist",
                    "html": "label.toggle",
                    "react": "Toggle",
                    "vue": "Toggle"
                }
            }
        )
        
        # ML-specific components
        self.add_category("ml", "Machine Learning Components")
        
        self.register_component(
            "model_prediction",
            "Model Prediction",
            "ml",
            {
                "description": "A component for making predictions with a machine learning model.",
                "properties": {
                    "model_source": {"type": "string", "required": True},
                    "feature_names": {"type": "array", "items": {"type": "string"}, "required": False},
                    "show_probabilities": {"type": "boolean", "default": True}
                },
                "frameworks": {
                    "streamlit": "custom",
                    "dash": "custom",
                    "html": "custom",
                    "react": "custom",
                    "vue": "custom"
                }
            }
        )
        
        self.register_component(
            "feature_importance",
            "Feature Importance",
            "ml",
            {
                "description": "A component for displaying feature importance of a machine learning model.",
                "properties": {
                    "model_source": {"type": "string", "required": True},
                    "top_n": {"type": "integer", "default": 10},
                    "orientation": {"type": "string", "enum": ["vertical", "horizontal"], "default": "horizontal"}
                },
                "frameworks": {
                    "streamlit": "custom",
                    "dash": "custom",
                    "html": "custom",
                    "react": "custom",
                    "vue": "custom"
                }
            }
        )
        
        self.register_component(
            "confusion_matrix",
            "Confusion Matrix",
            "ml",
            {
                "description": "A component for displaying a confusion matrix for classification models.",
                "properties": {
                    "data_source": {"type": "string", "required": True},
                    "true_labels": {"type": "string", "required": True},
                    "predicted_labels": {"type": "string", "required": True},
                    "normalize": {"type": "boolean", "default": False},
                    "color_map": {"type": "string", "default": "Blues"}
                },
                "frameworks": {
                    "streamlit": "custom",
                    "dash": "custom",
                    "html": "custom",
                    "react": "custom",
                    "vue": "custom"
                }
            }
        )
        
        self.register_component(
            "roc_curve",
            "ROC Curve",
            "ml",
            {
                "description": "A component for displaying a ROC curve for binary classification models.",
                "properties": {
                    "data_source": {"type": "string", "required": True},
                    "true_labels": {"type": "string", "required": True},
                    "predicted_probs": {"type": "string", "required": True},
                    "show_auc": {"type": "boolean", "default": True}
                },
                "frameworks": {
                    "streamlit": "custom",
                    "dash": "custom",
                    "html": "custom",
                    "react": "custom",
                    "vue": "custom"
                }
            }
        )
        
        self.register_component(
            "precision_recall_curve",
            "Precision-Recall Curve",
            "ml",
            {
                "description": "A component for displaying a precision-recall curve for binary classification models.",
                "properties": {
                    "data_source": {"type": "string", "required": True},
                    "true_labels": {"type": "string", "required": True},
                    "predicted_probs": {"type": "string", "required": True},
                    "show_auc": {"type": "boolean", "default": True}
                },
                "frameworks": {
                    "streamlit": "custom",
                    "dash": "custom",
                    "html": "custom",
                    "react": "custom",
                    "vue": "custom"
                }
            }
        )
        
        self.register_component(
            "residual_plot",
            "Residual Plot",
            "ml",
            {
                "description": "A component for displaying residuals of a regression model.",
                "properties": {
                    "data_source": {"type": "string", "required": True},
                    "true_values": {"type": "string", "required": True},
                    "predicted_values": {"type": "string", "required": True}
                },
                "frameworks": {
                    "streamlit": "custom",
                    "dash": "custom",
                    "html": "custom",
                    "react": "custom",
                    "vue": "custom"
                }
            }
        )
    
    def add_category(self, category_id: str, category_name: str) -> None:
        """
        Add a component category.
        
        Args:
            category_id: Unique identifier for the category.
            category_name: Display name for the category.
        """
        if category_id in self.categories:
            raise UIGenerationError(f"Category ID '{category_id}' already exists.")
        
        self.categories[category_id] = {
            "name": category_name,
            "components": []
        }
    
    def register_component(self, component_id: str, component_name: str,
                          category_id: str, config: Dict[str, Any]) -> None:
        """
        Register a component in the library.
        
        Args:
            component_id: Unique identifier for the component.
            component_name: Display name for the component.
            category_id: Category to which the component belongs.
            config: Configuration for the component.
        """
        if component_id in self.components:
            raise UIGenerationError(f"Component ID '{component_id}' already exists.")
        
        if category_id not in self.categories:
            raise UIGenerationError(f"Category ID '{category_id}' does not exist.")
        
        component = {
            "id": component_id,
            "name": component_name,
            "category": category_id,
            "config": config
        }
        
        self.components[component_id] = component
        self.categories[category_id]["components"].append(component_id)
    
    def get_component(self, component_id: str) -> Dict[str, Any]:
        """
        Get a component by ID.
        
        Args:
            component_id: ID of the component to get.
            
        Returns:
            Component configuration.
        """
        if component_id not in self.components:
            raise UIGenerationError(f"Component ID '{component_id}' does not exist.")
        
        return self.components[component_id]
    
    def get_components_by_category(self, category_id: str) -> List[Dict[str, Any]]:
        """
        Get all components in a category.
        
        Args:
            category_id: ID of the category.
            
        Returns:
            List of components in the category.
        """
        if category_id not in self.categories:
            raise UIGenerationError(f"Category ID '{category_id}' does not exist.")
        
        return [self.components[component_id] for component_id in self.categories[category_id]["components"]]
    
    def get_all_components(self) -> Dict[str, Dict[str, Any]]:
        """
        Get all components.
        
        Returns:
            Dictionary of all components.
        """
        return self.components
    
    def get_all_categories(self) -> Dict[str, Dict[str, Any]]:
        """
        Get all categories.
        
        Returns:
            Dictionary of all categories.
        """
        return self.categories
    
    def get_component_code(self, component_id: str, framework: str,
                          props: Dict[str, Any]) -> str:
        """
        Get the code for a component in a specific framework.
        
        Args:
            component_id: ID of the component.
            framework: Framework to generate code for.
            props: Properties for the component.
            
        Returns:
            Code for the component.
        """
        component = self.get_component(component_id)
        
        if framework not in component["config"]["frameworks"]:
            raise UIGenerationError(f"Framework '{framework}' not supported for component '{component_id}'.")
        
        # Validate required properties
        for prop_name, prop_config in component["config"]["properties"].items():
            if prop_config.get("required", False) and prop_name not in props:
                raise UIGenerationError(f"Required property '{prop_name}' missing for component '{component_id}'.")
        
        # Generate code based on framework
        if framework == "streamlit":
            return self._generate_streamlit_code(component, props)
        elif framework == "dash":
            return self._generate_dash_code(component, props)
        elif framework == "html":
            return self._generate_html_code(component, props)
        elif framework == "react":
            return self._generate_react_code(component, props)
        elif framework == "vue":
            return self._generate_vue_code(component, props)
        else:
            raise UIGenerationError(f"Unsupported framework: {framework}")
    
    def _generate_streamlit_code(self, component: Dict[str, Any], props: Dict[str, Any]) -> str:
        """
        Generate Streamlit code for a component.
        
        Args:
            component: Component configuration.
            props: Properties for the component.
            
        Returns:
            Streamlit code for the component.
        """
        component_type = component["id"]
        framework_func = component["config"]["frameworks"]["streamlit"]
        
        # Handle custom components
        if framework_func == "custom":
            if component_type == "model_prediction":
                return self._generate_streamlit_model_prediction(props)
            elif component_type == "feature_importance":
                return self._generate_streamlit_feature_importance(props)
            elif component_type == "confusion_matrix":
                return self._generate_streamlit_confusion_matrix(props)
            elif component_type == "roc_curve":
                return self._generate_streamlit_roc_curve(props)
            elif component_type == "precision_recall_curve":
                return self._generate_streamlit_precision_recall_curve(props)
            elif component_type == "residual_plot":
                return self._generate_streamlit_residual_plot(props)
            else:
                raise UIGenerationError(f"Custom component '{component_type}' not implemented for Streamlit.")
        
        # Generate code for standard components
        props_str = ", ".join([f"{k}={repr(v)}" for k, v in props.items()])
        return f"{framework_func}({props_str})"
    
    def _generate_dash_code(self, component: Dict[str, Any], props: Dict[str, Any]) -> str:
        """
        Generate Dash code for a component.
        
        Args:
            component: Component configuration.
            props: Properties for the component.
            
        Returns:
            Dash code for the component.
        """
        component_type = component["id"]
        framework_func = component["config"]["frameworks"]["dash"]
        
        # Handle custom components
        if framework_func == "custom":
            if component_type == "model_prediction":
                return self._generate_dash_model_prediction(props)
            elif component_type == "feature_importance":
                return self._generate_dash_feature_importance(props)
            elif component_type == "confusion_matrix":
                return self._generate_dash_confusion_matrix(props)
            elif component_type == "roc_curve":
                return self._generate_dash_roc_curve(props)
            elif component_type == "precision_recall_curve":
                return self._generate_dash_precision_recall_curve(props)
            elif component_type == "residual_plot":
                return self._generate_dash_residual_plot(props)
            else:
                raise UIGenerationError(f"Custom component '{component_type}' not implemented for Dash.")
        
        # Generate code for standard components
        props_str = ", ".join([f"{k}={repr(v)}" for k, v in props.items()])
        return f"{framework_func}({props_str})"
    
    def _generate_html_code(self, component: Dict[str, Any], props: Dict[str, Any]) -> str:
        """
        Generate HTML code for a component.
        
        Args:
            component: Component configuration.
            props: Properties for the component.
            
        Returns:
            HTML code for the component.
        """
        component_type = component["id"]
        framework_element = component["config"]["frameworks"]["html"]
        
        # Handle custom components
        if framework_element == "custom":
            if component_type == "model_prediction":
                return self._generate_html_model_prediction(props)
            elif component_type == "feature_importance":
                return self._generate_html_feature_importance(props)
            elif component_type == "confusion_matrix":
                return self._generate_html_confusion_matrix(props)
            elif component_type == "roc_curve":
                return self._generate_html_roc_curve(props)
            elif component_type == "precision_recall_curve":
                return self._generate_html_precision_recall_curve(props)
            elif component_type == "residual_plot":
                return self._generate_html_residual_plot(props)
            else:
                raise UIGenerationError(f"Custom component '{component_type}' not implemented for HTML.")
        
        # Parse element and class
        element_parts = framework_element.split(".")
        element_tag = element_parts[0]
        element_class = element_parts[1] if len(element_parts) > 1 else None
        
        # Generate attributes string
        attributes = []
        for k, v in props.items():
            if k == "text" or k == "content":
                continue
            
            if isinstance(v, bool):
                if v:
                    attributes.append(k)
            else:
                attributes.append(f'{k}="{v}"')
        
        if element_class:
            attributes.append(f'class="{element_class}"')
        
        attributes_str = " ".join(attributes)
        
        # Generate content
        content = ""
        if "text" in props:
            content = props["text"]
        elif "content" in props:
            content = props["content"]
        
        # Generate HTML
        if content:
            return f"<{element_tag} {attributes_str}>{content}</{element_tag}>"
        else:
            return f"<{element_tag} {attributes_str} />"
    
    def _generate_react_code(self, component: Dict[str, Any], props: Dict[str, Any]) -> str:
        """
        Generate React code for a component.
        
        Args:
            component: Component configuration.
            props: Properties for the component.
            
        Returns:
            React code for the component.
        """
        component_type = component["id"]
        framework_component = component["config"]["frameworks"]["react"]
        
        # Handle custom components
        if framework_component == "custom":
            if component_type == "model_prediction":
                return self._generate_react_model_prediction(props)
            elif component_type == "feature_importance":
                return self._generate_react_feature_importance(props)
            elif component_type == "confusion_matrix":
                return self._generate_react_confusion_matrix(props)
            elif component_type == "roc_curve":
                return self._generate_react_roc_curve(props)
            elif component_type == "precision_recall_curve":
                return self._generate_react_precision_recall_curve(props)
            elif component_type == "residual_plot":
                return self._generate_react_residual_plot(props)
            else:
                raise UIGenerationError(f"Custom component '{component_type}' not implemented for React.")
        
        # Generate props string
        props_parts = []
        for k, v in props.items():
            if isinstance(v, bool):
                if v:
                    props_parts.append(f"{k}")
                else:
                    props_parts.append(f"{k}={{{v}}}")
            elif isinstance(v, (int, float)):
                props_parts.append(f"{k}={{{v}}}")
            elif isinstance(v, str):
                props_parts.append(f'{k}="{v}"')
            else:
                props_parts.append(f"{k}={{{repr(v)}}}")
        
        props_str = " ".join(props_parts)
        
        # Generate React component
        return f"<{framework_component} {props_str} />"
    
    def _generate_vue_code(self, component: Dict[str, Any], props: Dict[str, Any]) -> str:
        """
        Generate Vue code for a component.
        
        Args:
            component: Component configuration.
            props: Properties for the component.
            
        Returns:
            Vue code for the component.
        """
        component_type = component["id"]
        framework_component = component["config"]["frameworks"]["vue"]
        
        # Handle custom components
        if framework_component == "custom":
            if component_type == "model_prediction":
                return self._generate_vue_model_prediction(props)
            elif component_type == "feature_importance":
                return self._generate_vue_feature_importance(props)
            elif component_type == "confusion_matrix":
                return self._generate_vue_confusion_matrix(props)
            elif component_type == "roc_curve":
                return self._generate_vue_roc_curve(props)
            elif component_type == "precision_recall_curve":
                return self._generate_vue_precision_recall_curve(props)
            elif component_type == "residual_plot":
                return self._generate_vue_residual_plot(props)
            else:
                raise UIGenerationError(f"Custom component '{component_type}' not implemented for Vue.")
        
        # Generate props string
        props_parts = []
        for k, v in props.items():
            if isinstance(v, bool):
                if v:
                    props_parts.append(f"{k}")
                else:
                    props_parts.append(f":{k}=\"{str(v).lower()}\"")
            elif isinstance(v, (int, float)):
                props_parts.append(f":{k}=\"{v}\"")
            elif isinstance(v, str):
                props_parts.append(f'{k}="{v}"')
            else:
                props_parts.append(f":{k}=\"{repr(v)}\"")
        
        props_str = " ".join(props_parts)
        
        # Generate Vue component
        return f"<{framework_component} {props_str} />"
    
    # Custom component generators for Streamlit
    def _generate_streamlit_model_prediction(self, props: Dict[str, Any]) -> str:
        """
        Generate Streamlit code for model prediction component.
        
        Args:
            props: Properties for the component.
            
        Returns:
            Streamlit code for the component.
        """
        model_source = props["model_source"]
        feature_names = props.get("feature_names", [])
        show_probabilities = props.get("show_probabilities", True)
        
        code = f"""
# Model Prediction Component
st.subheader("Model Prediction")

# Get model
model = data_sources["{model_source}"]

# Create input fields
input_values = {{}}
"""
        
        if feature_names:
            code += """
# Use provided feature names
feature_names = {feature_names}
""".format(feature_names=repr(feature_names))
        else:
            code += """
# Try to get feature names from model
if hasattr(model, "feature_names_in_"):
    feature_names = model.feature_names_in_
else:
    st.error("No feature names provided and model does not have feature_names_in_ attribute.")
    feature_names = []
"""
        
        code += """
# Create input fields for each feature
for feature in feature_names:
    input_values[feature] = st.number_input(f"{feature}")

# Add prediction button
if st.button("Predict"):
    try:
        # Create input DataFrame
        input_df = pd.DataFrame([input_values])
        
        # Make prediction
        prediction = model.predict(input_df)
        
        st.success(f"Prediction: {prediction[0]}")
"""
        
        if show_probabilities:
            code += """        
        # Show probability if available
        if hasattr(model, "predict_proba"):
            probabilities = model.predict_proba(input_df)
            st.write("Prediction probabilities:")
            
            # Get class names if available
            if hasattr(model, "classes_"):
                class_names = model.classes_
                prob_df = pd.DataFrame([probabilities[0]], columns=class_names)
            else:
                prob_df = pd.DataFrame([probabilities[0]], columns=[f"Class {i}" for i in range(probabilities.shape[1])])
            
            st.dataframe(prob_df)
"""
        
        code += """
    except Exception as e:
        st.error(f"Error making prediction: {str(e)}")
"""
        
        return code
    
    def _generate_streamlit_feature_importance(self, props: Dict[str, Any]) -> str:
        """
        Generate Streamlit code for feature importance component.
        
        Args:
            props: Properties for the component.
            
        Returns:
            Streamlit code for the component.
        """
        model_source = props["model_source"]
        top_n = props.get("top_n", 10)
        orientation = props.get("orientation", "horizontal")
        
        code = f"""
# Feature Importance Component
st.subheader("Feature Importance")

# Get model
model = data_sources["{model_source}"]

try:
    # Get feature importance
    if hasattr(model, "feature_importances_"):
        importances = model.feature_importances_
    elif hasattr(model, "coef_"):
        importances = np.abs(model.coef_[0]) if len(model.coef_.shape) > 1 else np.abs(model.coef_)
    else:
        st.error("Model does not have feature_importances_ or coef_ attribute.")
        importances = []
    
    # Get feature names
    if hasattr(model, "feature_names_in_"):
        feature_names = model.feature_names_in_
    else:
        feature_names = [f"Feature {{i}}" for i in range(len(importances))]
    
    # Create DataFrame
    importance_df = pd.DataFrame({{"Feature": feature_names, "Importance": importances}})
    importance_df = importance_df.sort_values("Importance", ascending=False).head({top_n})
    
    # Plot feature importance
    fig, ax = plt.subplots(figsize=(10, 6))
    
    if "{orientation}" == "horizontal":
        sns.barplot(x="Importance", y="Feature", data=importance_df, ax=ax)
        ax.set_xlabel("Importance")
        ax.set_ylabel("Feature")
    else:
        sns.barplot(x="Feature", y="Importance", data=importance_df, ax=ax)
        ax.set_xlabel("Feature")
        ax.set_ylabel("Importance")
        plt.xticks(rotation=45, ha="right")
    
    ax.set_title("Feature Importance")
    plt.tight_layout()
    
    st.pyplot(fig)
    
    # Show importance table
    st.dataframe(importance_df)
except Exception as e:
    st.error(f"Error displaying feature importance: {{str(e)}}")
"""
        
        return code
    
    def _generate_streamlit_confusion_matrix(self, props: Dict[str, Any]) -> str:
        """
        Generate Streamlit code for confusion matrix component.
        
        Args:
            props: Properties for the component.
            
        Returns:
            Streamlit code for the component.
        """
        data_source = props["data_source"]
        true_labels = props["true_labels"]
        predicted_labels = props["predicted_labels"]
        normalize = props.get("normalize", False)
        color_map = props.get("color_map", "Blues")
        
        code = f"""
# Confusion Matrix Component
st.subheader("Confusion Matrix")

# Get data
data = data_sources["{data_source}"]
y_true = data["{true_labels}"]
y_pred = data["{predicted_labels}"]

try:
    from sklearn.metrics import confusion_matrix
    import seaborn as sns
    
    # Compute confusion matrix
    cm = confusion_matrix(y_true, y_pred)
    
    # Normalize if requested
    if {normalize}:
        cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]
    
    # Get class names if available
    if hasattr(data, "classes_"):
        class_names = data.classes_
    else:
        class_names = sorted(set(y_true) | set(y_pred))
    
    # Plot confusion matrix
    fig, ax = plt.subplots(figsize=(10, 8))
    sns.heatmap(cm, annot=True, fmt='.2f' if {normalize} else 'd', cmap='{color_map}',
                xticklabels=class_names, yticklabels=class_names, ax=ax)
    ax.set_xlabel('Predicted')
    ax.set_ylabel('True')
    ax.set_title('Confusion Matrix')
    plt.tight_layout()
    
    st.pyplot(fig)
except Exception as e:
    st.error(f"Error displaying confusion matrix: {{str(e)}}")
"""
        
        return code
    
    def _generate_streamlit_roc_curve(self, props: Dict[str, Any]) -> str:
        """
        Generate Streamlit code for ROC curve component.
        
        Args:
            props: Properties for the component.
            
        Returns:
            Streamlit code for the component.
        """
        data_source = props["data_source"]
        true_labels = props["true_labels"]
        predicted_probs = props["predicted_probs"]
        show_auc = props.get("show_auc", True)
        
        code = f"""
# ROC Curve Component
st.subheader("ROC Curve")

# Get data
data = data_sources["{data_source}"]
y_true = data["{true_labels}"]
y_score = data["{predicted_probs}"]

try:
    from sklearn.metrics import roc_curve, auc
    
    # Compute ROC curve and ROC area
    fpr, tpr, _ = roc_curve(y_true, y_score)
    roc_auc = auc(fpr, tpr)
    
    # Plot ROC curve
    fig, ax = plt.subplots(figsize=(10, 8))
    ax.plot(fpr, tpr, label=f'ROC curve (area = {{roc_auc:.2f}})' if {show_auc} else 'ROC curve')
    ax.plot([0, 1], [0, 1], 'k--')
    ax.set_xlim([0.0, 1.0])
    ax.set_ylim([0.0, 1.05])
    ax.set_xlabel('False Positive Rate')
    ax.set_ylabel('True Positive Rate')
    ax.set_title('Receiver Operating Characteristic (ROC)')
    ax.legend(loc="lower right")
    plt.tight_layout()
    
    st.pyplot(fig)
    
    # Show AUC value
    if {show_auc}:
        st.metric("Area Under the Curve (AUC)", f"{{roc_auc:.4f}}")
except Exception as e:
    st.error(f"Error displaying ROC curve: {{str(e)}}")
"""
        
        return code
    
    def _generate_streamlit_precision_recall_curve(self, props: Dict[str, Any]) -> str:
        """
        Generate Streamlit code for precision-recall curve component.
        
        Args:
            props: Properties for the component.
            
        Returns:
            Streamlit code for the component.
        """
        data_source = props["data_source"]
        true_labels = props["true_labels"]
        predicted_probs = props["predicted_probs"]
        show_auc = props.get("show_auc", True)
        
        code = f"""
# Precision-Recall Curve Component
st.subheader("Precision-Recall Curve")

# Get data
data = data_sources["{data_source}"]
y_true = data["{true_labels}"]
y_score = data["{predicted_probs}"]

try:
    from sklearn.metrics import precision_recall_curve, average_precision_score
    
    # Compute Precision-Recall curve and average precision
    precision, recall, _ = precision_recall_curve(y_true, y_score)
    ap = average_precision_score(y_true, y_score)
    
    # Plot Precision-Recall curve
    fig, ax = plt.subplots(figsize=(10, 8))
    ax.plot(recall, precision, label=f'Precision-Recall curve (AP = {{ap:.2f}})' if {show_auc} else 'Precision-Recall curve')
    ax.set_xlim([0.0, 1.0])
    ax.set_ylim([0.0, 1.05])
    ax.set_xlabel('Recall')
    ax.set_ylabel('Precision')
    ax.set_title('Precision-Recall Curve')
    ax.legend(loc="lower left")
    plt.tight_layout()
    
    st.pyplot(fig)
    
    # Show AP value
    if {show_auc}:
        st.metric("Average Precision (AP)", f"{{ap:.4f}}")
except Exception as e:
    st.error(f"Error displaying Precision-Recall curve: {{str(e)}}")
"""
        
        return code
    
    def _generate_streamlit_residual_plot(self, props: Dict[str, Any]) -> str:
        """
        Generate Streamlit code for residual plot component.
        
        Args:
            props: Properties for the component.
            
        Returns:
            Streamlit code for the component.
        """
        data_source = props["data_source"]
        true_values = props["true_values"]
        predicted_values = props["predicted_values"]
        
        code = f"""
# Residual Plot Component
st.subheader("Residual Plot")

# Get data
data = data_sources["{data_source}"]
y_true = data["{true_values}"]
y_pred = data["{predicted_values}"]

try:
    # Compute residuals
    residuals = y_true - y_pred
    
    # Create figure with subplots
    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(15, 6))
    
    # Plot residuals vs predicted values
    ax1.scatter(y_pred, residuals)
    ax1.axhline(y=0, color='r', linestyle='-')
    ax1.set_xlabel('Predicted Values')
    ax1.set_ylabel('Residuals')
    ax1.set_title('Residuals vs Predicted Values')
    
    # Plot histogram of residuals
    ax2.hist(residuals, bins=20, edgecolor='black')
    ax2.axvline(x=0, color='r', linestyle='-')
    ax2.set_xlabel('Residuals')
    ax2.set_ylabel('Frequency')
    ax2.set_title('Histogram of Residuals')
    
    plt.tight_layout()
    
    st.pyplot(fig)
    
    # Show residual statistics
    st.write("Residual Statistics:")
    stats_df = pd.DataFrame({{
        "Statistic": ["Mean", "Std Dev", "Min", "25%", "Median", "75%", "Max"],
        "Value": [
            residuals.mean(),
            residuals.std(),
            residuals.min(),
            np.percentile(residuals, 25),
            np.median(residuals),
            np.percentile(residuals, 75),
            residuals.max()
        ]
    }})
    st.dataframe(stats_df)
except Exception as e:
    st.error(f"Error displaying residual plot: {{str(e)}}")
"""
        
        return code
    
    # Custom component generators for other frameworks would be implemented similarly
    def _generate_dash_model_prediction(self, props: Dict[str, Any]) -> str:
        """Placeholder for Dash model prediction component"""
        return "# Dash model prediction component (implementation details omitted for brevity)"
    
    def _generate_dash_feature_importance(self, props: Dict[str, Any]) -> str:
        """Placeholder for Dash feature importance component"""
        return "# Dash feature importance component (implementation details omitted for brevity)"
    
    def _generate_dash_confusion_matrix(self, props: Dict[str, Any]) -> str:
        """Placeholder for Dash confusion matrix component"""
        return "# Dash confusion matrix component (implementation details omitted for brevity)"
    
    def _generate_dash_roc_curve(self, props: Dict[str, Any]) -> str:
        """Placeholder for Dash ROC curve component"""
        return "# Dash ROC curve component (implementation details omitted for brevity)"
    
    def _generate_dash_precision_recall_curve(self, props: Dict[str, Any]) -> str:
        """Placeholder for Dash precision-recall curve component"""
        return "# Dash precision-recall curve component (implementation details omitted for brevity)"
    
    def _generate_dash_residual_plot(self, props: Dict[str, Any]) -> str:
        """Placeholder for Dash residual plot component"""
        return "# Dash residual plot component (implementation details omitted for brevity)"
    
    def _generate_html_model_prediction(self, props: Dict[str, Any]) -> str:
        """Placeholder for HTML model prediction component"""
        return "<!-- HTML model prediction component (implementation details omitted for brevity) -->"
    
    def _generate_html_feature_importance(self, props: Dict[str, Any]) -> str:
        """Placeholder for HTML feature importance component"""
        return "<!-- HTML feature importance component (implementation details omitted for brevity) -->"
    
    def _generate_html_confusion_matrix(self, props: Dict[str, Any]) -> str:
        """Placeholder for HTML confusion matrix component"""
        return "<!-- HTML confusion matrix component (implementation details omitted for brevity) -->"
    
    def _generate_html_roc_curve(self, props: Dict[str, Any]) -> str:
        """Placeholder for HTML ROC curve component"""
        return "<!-- HTML ROC curve component (implementation details omitted for brevity) -->"
    
    def _generate_html_precision_recall_curve(self, props: Dict[str, Any]) -> str:
        """Placeholder for HTML precision-recall curve component"""
        return "<!-- HTML precision-recall curve component (implementation details omitted for brevity) -->"
    
    def _generate_html_residual_plot(self, props: Dict[str, Any]) -> str:
        """Placeholder for HTML residual plot component"""
        return "<!-- HTML residual plot component (implementation details omitted for brevity) -->"
    
    def _generate_react_model_prediction(self, props: Dict[str, Any]) -> str:
        """Placeholder for React model prediction component"""
        return "// React model prediction component (implementation details omitted for brevity)"
    
    def _generate_react_feature_importance(self, props: Dict[str, Any]) -> str:
        """Placeholder for React feature importance component"""
        return "// React feature importance component (implementation details omitted for brevity)"
    
    def _generate_react_confusion_matrix(self, props: Dict[str, Any]) -> str:
        """Placeholder for React confusion matrix component"""
        return "// React confusion matrix component (implementation details omitted for brevity)"
    
    def _generate_react_roc_curve(self, props: Dict[str, Any]) -> str:
        """Placeholder for React ROC curve component"""
        return "// React ROC curve component (implementation details omitted for brevity)"
    
    def _generate_react_precision_recall_curve(self, props: Dict[str, Any]) -> str:
        """Placeholder for React precision-recall curve component"""
        return "// React precision-recall curve component (implementation details omitted for brevity)"
    
    def _generate_react_residual_plot(self, props: Dict[str, Any]) -> str:
        """Placeholder for React residual plot component"""
        return "// React residual plot component (implementation details omitted for brevity)"
    
    def _generate_vue_model_prediction(self, props: Dict[str, Any]) -> str:
        """Placeholder for Vue model prediction component"""
        return "<!-- Vue model prediction component (implementation details omitted for brevity) -->"
    
    def _generate_vue_feature_importance(self, props: Dict[str, Any]) -> str:
        """Placeholder for Vue feature importance component"""
        return "<!-- Vue feature importance component (implementation details omitted for brevity) -->"
    
    def _generate_vue_confusion_matrix(self, props: Dict[str, Any]) -> str:
        """Placeholder for Vue confusion matrix component"""
        return "<!-- Vue confusion matrix component (implementation details omitted for brevity) -->"
    
    def _generate_vue_roc_curve(self, props: Dict[str, Any]) -> str:
        """Placeholder for Vue ROC curve component"""
        return "<!-- Vue ROC curve component (implementation details omitted for brevity) -->"
    
    def _generate_vue_precision_recall_curve(self, props: Dict[str, Any]) -> str:
        """Placeholder for Vue precision-recall curve component"""
        return "<!-- Vue precision-recall curve component (implementation details omitted for brevity) -->"
    
    def _generate_vue_residual_plot(self, props: Dict[str, Any]) -> str:
        """Placeholder for Vue residual plot component"""
        return "<!-- Vue residual plot component (implementation details omitted for brevity) -->"



"""
Hook system for SBYB plugins.

This module provides functionality for registering and executing hooks,
which allow plugins to extend or modify the behavior of SBYB components.
"""

from typing import Any, Dict, List, Optional, Union, Tuple, Callable
import inspect
import logging
from collections import defaultdict

from sbyb.core.base import SBYBComponent
from sbyb.core.exceptions import HookError


class HookManager(SBYBComponent):
    """
    Hook manager for SBYB.
    
    This component provides functionality for registering and executing hooks,
    which allow plugins to extend or modify the behavior of SBYB components.
    """
    
    def __init__(self, config: Optional[Dict[str, Any]] = None):
        """
        Initialize the hook manager.
        
        Args:
            config: Configuration dictionary for the hook manager.
        """
        super().__init__(config)
        
        # Set up logging
        self.logger = logging.getLogger("sbyb.plugins.hooks")
        
        # Initialize hooks registry
        self.hooks = defaultdict(list)
    
    def register_hook(self, hook_name: str, callback: Callable, priority: int = 100) -> None:
        """
        Register a hook.
        
        Args:
            hook_name: Name of the hook.
            callback: Callback function to execute when the hook is triggered.
            priority: Priority of the hook (lower numbers run first).
        """
        self.hooks[hook_name].append({
            "callback": callback,
            "priority": priority
        })
        
        # Sort hooks by priority
        self.hooks[hook_name] = sorted(self.hooks[hook_name], key=lambda x: x["priority"])
        
        self.logger.debug(f"Registered hook {hook_name} with priority {priority}")
    
    def register_hooks_from_instance(self, instance: Any) -> None:
        """
        Register all hooks from an instance.
        
        Args:
            instance: Instance to register hooks from.
        """
        for name, method in inspect.getmembers(instance, inspect.ismethod):
            if hasattr(method, "__hook_info__"):
                hook_info = getattr(method, "__hook_info__")
                hook_name = hook_info.get("name")
                priority = hook_info.get("priority", 100)
                
                self.register_hook(hook_name, method, priority)
    
    def unregister_hook(self, hook_name: str, callback: Optional[Callable] = None) -> None:
        """
        Unregister a hook.
        
        Args:
            hook_name: Name of the hook.
            callback: Callback function to unregister. If None, unregister all hooks with the given name.
        """
        if hook_name not in self.hooks:
            return
        
        if callback is None:
            # Unregister all hooks with the given name
            self.hooks[hook_name] = []
            self.logger.debug(f"Unregistered all hooks for {hook_name}")
        else:
            # Unregister specific callback
            self.hooks[hook_name] = [h for h in self.hooks[hook_name] if h["callback"] != callback]
            self.logger.debug(f"Unregistered hook {hook_name} for specific callback")
    
    def has_hooks(self, hook_name: str) -> bool:
        """
        Check if a hook has any registered callbacks.
        
        Args:
            hook_name: Name of the hook.
            
        Returns:
            True if the hook has registered callbacks, False otherwise.
        """
        return hook_name in self.hooks and len(self.hooks[hook_name]) > 0
    
    def execute_hook(self, hook_name: str, *args, **kwargs) -> List[Any]:
        """
        Execute a hook.
        
        Args:
            hook_name: Name of the hook.
            *args: Positional arguments to pass to the hook callbacks.
            **kwargs: Keyword arguments to pass to the hook callbacks.
            
        Returns:
            List of results from all hook callbacks.
        """
        if hook_name not in self.hooks:
            return []
        
        results = []
        
        for hook in self.hooks[hook_name]:
            try:
                result = hook["callback"](*args, **kwargs)
                results.append(result)
            except Exception as e:
                self.logger.error(f"Error executing hook {hook_name}: {str(e)}")
                results.append(None)
        
        return results
    
    def execute_hook_until_true(self, hook_name: str, *args, **kwargs) -> Tuple[bool, Any]:
        """
        Execute a hook until a callback returns a truthy value.
        
        Args:
            hook_name: Name of the hook.
            *args: Positional arguments to pass to the hook callbacks.
            **kwargs: Keyword arguments to pass to the hook callbacks.
            
        Returns:
            Tuple of (success, result) where success is True if any callback returned a truthy value,
            and result is the value returned by that callback.
        """
        if hook_name not in self.hooks:
            return False, None
        
        for hook in self.hooks[hook_name]:
            try:
                result = hook["callback"](*args, **kwargs)
                if result:
                    return True, result
            except Exception as e:
                self.logger.error(f"Error executing hook {hook_name}: {str(e)}")
        
        return False, None
    
    def execute_hook_until_false(self, hook_name: str, *args, **kwargs) -> Tuple[bool, Any]:
        """
        Execute a hook until a callback returns a falsy value.
        
        Args:
            hook_name: Name of the hook.
            *args: Positional arguments to pass to the hook callbacks.
            **kwargs: Keyword arguments to pass to the hook callbacks.
            
        Returns:
            Tuple of (success, result) where success is True if all callbacks returned truthy values,
            and result is the value returned by the last callback.
        """
        if hook_name not in self.hooks:
            return True, None
        
        last_result = None
        
        for hook in self.hooks[hook_name]:
            try:
                result = hook["callback"](*args, **kwargs)
                if not result:
                    return False, result
                last_result = result
            except Exception as e:
                self.logger.error(f"Error executing hook {hook_name}: {str(e)}")
                return False, None
        
        return True, last_result
    
    def execute_hook_waterfall(self, hook_name: str, initial_value: Any, *args, **kwargs) -> Any:
        """
        Execute a hook in waterfall mode, passing the result of each callback to the next.
        
        Args:
            hook_name: Name of the hook.
            initial_value: Initial value to pass to the first callback.
            *args: Positional arguments to pass to the hook callbacks.
            **kwargs: Keyword arguments to pass to the hook callbacks.
            
        Returns:
            Final result after all callbacks have been executed.
        """
        if hook_name not in self.hooks:
            return initial_value
        
        result = initial_value
        
        for hook in self.hooks[hook_name]:
            try:
                result = hook["callback"](result, *args, **kwargs)
            except Exception as e:
                self.logger.error(f"Error executing hook {hook_name}: {str(e)}")
        
        return result
    
    def execute_hook_parallel(self, hook_name: str, *args, **kwargs) -> List[Any]:
        """
        Execute a hook in parallel mode, running all callbacks concurrently.
        
        Args:
            hook_name: Name of the hook.
            *args: Positional arguments to pass to the hook callbacks.
            **kwargs: Keyword arguments to pass to the hook callbacks.
            
        Returns:
            List of results from all hook callbacks.
        """
        if hook_name not in self.hooks:
            return []
        
        import concurrent.futures
        
        results = []
        
        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = []
            
            for hook in self.hooks[hook_name]:
                future = executor.submit(hook["callback"], *args, **kwargs)
                futures.append(future)
            
            for future in concurrent.futures.as_completed(futures):
                try:
                    result = future.result()
                    results.append(result)
                except Exception as e:
                    self.logger.error(f"Error executing hook {hook_name}: {str(e)}")
                    results.append(None)
        
        return results



"""
Model stacking component for SBYB AutoML.

This module provides functionality for creating stacked ensemble models
by combining multiple base models.
"""

from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, clone
from sklearn.model_selection import KFold, StratifiedKFold
from sklearn.linear_model import LogisticRegression, Ridge
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor

from sbyb.core.base import SBYBComponent
from sbyb.core.exceptions import StackingError


class ModelStacker(SBYBComponent):
    """
    Model stacking component.
    
    This component creates stacked ensemble models by combining multiple base models.
    """
    
    def __init__(self, meta_learner: Optional[str] = 'auto', 
                 cv_folds: int = 5, use_features: bool = True,
                 config: Optional[Dict[str, Any]] = None):
        """
        Initialize the model stacker.
        
        Args:
            meta_learner: Meta-learner model type ('auto', 'linear', 'forest').
            cv_folds: Number of cross-validation folds for generating meta-features.
            use_features: Whether to include original features in the meta-learner.
            config: Configuration dictionary for the stacker.
        """
        super().__init__(config)
        
        self.meta_learner = meta_learner
        self.cv_folds = cv_folds
        self.use_features = use_features
        self.base_models = None
        self.meta_model = None
    
    def stack_models(self, models: List[BaseEstimator], X_train: pd.DataFrame, y: Union[pd.Series, np.ndarray],
                    X_val: Optional[pd.DataFrame] = None, y_val: Optional[Union[pd.Series, np.ndarray]] = None,
                    task: str = 'auto', time_limit: Optional[float] = None) -> BaseEstimator:
        """
        Create a stacked ensemble model.
        
        Args:
            models: List of base models to stack.
            X_train: Training features.
            y: Target variable.
            X_val: Validation features (optional).
            y_val: Validation target (optional).
            task: Machine learning task type ('auto', 'binary', 'multiclass', 'regression').
            time_limit: Time limit in seconds for stacking.
            
        Returns:
            Stacked ensemble model.
        """
        import time
        start_time = time.time()
        
        # Store base models
        self.base_models = models
        
        # Determine task type if auto
        if task == 'auto':
            task = self._detect_task(y)
        
        # Generate meta-features using cross-validation
        meta_features_train = self._generate_meta_features(models, X_train, y, task)
        
        # Generate meta-features for validation set if provided
        meta_features_val = None
        if X_val is not None and y_val is not None:
            # Fit base models on all training data
            fitted_models = []
            for model in models:
                fitted_model = clone(model)
                fitted_model.fit(X_train, y)
                fitted_models.append(fitted_model)
            
            # Generate predictions for validation set
            meta_features_val = np.column_stack([
                model.predict_proba(X_val)[:, 1] if task == 'binary' and hasattr(model, 'predict_proba')
                else model.predict_proba(X_val) if task == 'multiclass' and hasattr(model, 'predict_proba')
                else model.predict(X_val).reshape(-1, 1)
                for model in fitted_models
            ])
            
            # Include original features if specified
            if self.use_features:
                meta_features_val = np.column_stack([meta_features_val, X_val.values])
        
        # Create meta-learner
        meta_model = self._create_meta_learner(task)
        
        # Fit meta-learner
        meta_model.fit(meta_features_train, y)
        
        # Create and return stacked model
        stacked_model = StackedEnsemble(
            base_models=models,
            meta_model=meta_model,
            task=task,
            use_features=self.use_features
        )
        
        # Store meta-model
        self.meta_model = meta_model
        
        return stacked_model
    
    def _generate_meta_features(self, models: List[BaseEstimator], X: pd.DataFrame, 
                               y: Union[pd.Series, np.ndarray], task: str) -> np.ndarray:
        """
        Generate meta-features using cross-validation.
        
        Args:
            models: List of base models.
            X: Input features.
            y: Target variable.
            task: Machine learning task type.
            
        Returns:
            Meta-features array.
        """
        # Create cross-validation object
        if task in ['binary', 'multiclass']:
            cv = StratifiedKFold(n_splits=self.cv_folds, shuffle=True, random_state=42)
        else:
            cv = KFold(n_splits=self.cv_folds, shuffle=True, random_state=42)
        
        # Initialize meta-features array
        n_models = len(models)
        meta_features = np.zeros((X.shape[0], n_models))
        
        # Generate meta-features for each fold
        for train_idx, val_idx in cv.split(X, y):
            X_fold_train, X_fold_val = X.iloc[train_idx], X.iloc[val_idx]
            y_fold_train = y.iloc[train_idx] if isinstance(y, pd.Series) else y[train_idx]
            
            # Train each model on the fold
            for i, model in enumerate(models):
                # Clone the model to avoid modifying the original
                model_clone = clone(model)
                
                try:
                    # Fit the model
                    model_clone.fit(X_fold_train, y_fold_train)
                    
                    # Generate predictions
                    if task == 'binary' and hasattr(model_clone, 'predict_proba'):
                        # For binary classification with probability support
                        meta_features[val_idx, i] = model_clone.predict_proba(X_fold_val)[:, 1]
                    elif task == 'multiclass' and hasattr(model_clone, 'predict_proba'):
                        # For multiclass, use the predicted class (not probabilities)
                        # This is a simplification; in practice, you might want to use all class probabilities
                        meta_features[val_idx, i] = model_clone.predict(X_fold_val)
                    else:
                        # For regression or models without probability support
                        meta_features[val_idx, i] = model_clone.predict(X_fold_val)
                except Exception as e:
                    # If model fails, use zeros
                    meta_features[val_idx, i] = 0
                    print(f"Error in model {i} for fold: {str(e)}")
        
        # Include original features if specified
        if self.use_features:
            meta_features = np.column_stack([meta_features, X.values])
        
        return meta_features
    
    def _create_meta_learner(self, task: str) -> BaseEstimator:
        """
        Create a meta-learner model.
        
        Args:
            task: Machine learning task type.
            
        Returns:
            Meta-learner model.
        """
        if self.meta_learner == 'auto':
            # Choose meta-learner based on task
            if task == 'binary':
                return LogisticRegression(C=1.0, solver='lbfgs', max_iter=1000)
            elif task == 'multiclass':
                return RandomForestClassifier(n_estimators=100, random_state=42)
            else:  # regression
                return Ridge(alpha=1.0)
        
        elif self.meta_learner == 'linear':
            # Use linear models
            if task in ['binary', 'multiclass']:
                return LogisticRegression(C=1.0, solver='lbfgs', max_iter=1000)
            else:  # regression
                return Ridge(alpha=1.0)
        
        elif self.meta_learner == 'forest':
            # Use random forest
            if task in ['binary', 'multiclass']:
                return RandomForestClassifier(n_estimators=100, random_state=42)
            else:  # regression
                return RandomForestRegressor(n_estimators=100, random_state=42)
        
        else:
            raise StackingError(f"Unknown meta-learner type: {self.meta_learner}")
    
    def _detect_task(self, y: Union[pd.Series, np.ndarray]) -> str:
        """
        Detect the machine learning task type from the target variable.
        
        Args:
            y: Target variable.
            
        Returns:
            Task type ('binary', 'multiclass', or 'regression').
        """
        if isinstance(y, np.ndarray):
            y = pd.Series(y)
        
        # Check if it's classification or regression
        if pd.api.types.is_numeric_dtype(y.dtype):
            n_unique = y.nunique()
            
            if n_unique == 2:
                return 'binary'
            elif n_unique <= 20:  # Arbitrary threshold for multiclass
                return 'multiclass'
            else:
                return 'regression'
        else:
            # Non-numeric types are assumed to be classification
            return 'multiclass'


class StackedEnsemble(BaseEstimator):
    """
    Stacked ensemble model.
    
    This class represents a stacked ensemble model that combines multiple base models
    with a meta-learner.
    """
    
    def __init__(self, base_models: List[BaseEstimator], meta_model: BaseEstimator, 
                task: str, use_features: bool = True):
        """
        Initialize the stacked ensemble model.
        
        Args:
            base_models: List of base models.
            meta_model: Meta-learner model.
            task: Machine learning task type.
            use_features: Whether to include original features in the meta-learner.
        """
        self.base_models = base_models
        self.meta_model = meta_model
        self.task = task
        self.use_features = use_features
        self._is_fitted = False
    
    def fit(self, X: pd.DataFrame, y: Union[pd.Series, np.ndarray]) -> 'StackedEnsemble':
        """
        Fit the stacked ensemble model.
        
        Args:
            X: Input features.
            y: Target variable.
            
        Returns:
            self: The fitted model.
        """
        # Fit base models
        for model in self.base_models:
            model.fit(X, y)
        
        # Generate meta-features
        meta_features = self._generate_meta_features(X)
        
        # Fit meta-model
        self.meta_model.fit(meta_features, y)
        
        self._is_fitted = True
        return self
    
    def predict(self, X: pd.DataFrame) -> np.ndarray:
        """
        Make predictions using the stacked ensemble model.
        
        Args:
            X: Input features.
            
        Returns:
            Predictions.
        """
        if not self._is_fitted:
            raise StackingError("Model has not been fitted yet. Call 'fit' before using 'predict'.")
        
        # Generate meta-features
        meta_features = self._generate_meta_features(X)
        
        # Make predictions with meta-model
        return self.meta_model.predict(meta_features)
    
    def predict_proba(self, X: pd.DataFrame) -> np.ndarray:
        """
        Predict class probabilities using the stacked ensemble model.
        
        Args:
            X: Input features.
            
        Returns:
            Class probabilities.
        """
        if not self._is_fitted:
            raise StackingError("Model has not been fitted yet. Call 'fit' before using 'predict_proba'.")
        
        if self.task not in ['binary', 'multiclass']:
            raise StackingError("predict_proba is only available for classification tasks")
        
        # Check if meta-model supports probability predictions
        if not hasattr(self.meta_model, 'predict_proba'):
            raise StackingError("Meta-model does not support probability predictions")
        
        # Generate meta-features
        meta_features = self._generate_meta_features(X)
        
        # Make probability predictions with meta-model
        return self.meta_model.predict_proba(meta_features)
    
    def _generate_meta_features(self, X: pd.DataFrame) -> np.ndarray:
        """
        Generate meta-features from base models.
        
        Args:
            X: Input features.
            
        Returns:
            Meta-features array.
        """
        # Generate predictions from base models
        meta_features = np.column_stack([
            model.predict_proba(X)[:, 1] if self.task == 'binary' and hasattr(model, 'predict_proba')
            else model.predict_proba(X) if self.task == 'multiclass' and hasattr(model, 'predict_proba')
            else model.predict(X).reshape(-1, 1)
            for model in self.base_models
        ])
        
        # Include original features if specified
        if self.use_features:
            meta_features = np.column_stack([meta_features, X.values])
        
        return meta_features



"""
Run class for SBYB tracking.

This module provides the Run class, which represents a single execution of a machine
learning workflow with specific parameters, metrics, and artifacts.
"""

from typing import Any, Dict, List, Optional, Union
import os
import json
import datetime
import uuid
from pathlib import Path
import pickle
import matplotlib.pyplot as plt
import numpy as np

from sbyb.core.base import SBYBComponent
from sbyb.core.exceptions import TrackingError


class Run(SBYBComponent):
    """
    Represents a single execution of a machine learning workflow.
    
    A run tracks parameters, metrics, artifacts, and other metadata for a single
    execution of a machine learning workflow. It provides methods for logging
    metrics, parameters, artifacts, and other information about the run.
    """
    
    def __init__(self, name: Optional[str] = None, description: Optional[str] = None,
                tags: Optional[List[str]] = None, config: Optional[Dict[str, Any]] = None):
        """
        Initialize a new run.
        
        Args:
            name: Optional name of the run.
            description: Optional description of the run.
            tags: Optional list of tags for categorizing the run.
            config: Optional configuration dictionary.
        """
        super().__init__(config)
        
        self.name = name or f"Run_{datetime.datetime.now().strftime('%Y%m%d_%H%M%S')}"
        self.description = description or ""
        self.tags = tags or []
        self.run_id = str(uuid.uuid4())
        self.experiment_id = None
        self.created_at = datetime.datetime.now().isoformat()
        self.started_at = None
        self.ended_at = None
        self.status = "created"  # created, running, completed, failed
        self.parameters = {}
        self.metrics = {}
        self.artifacts = {}
        self.logs = []
        self.metadata = {}
        self.history = {}
    
    def start(self) -> None:
        """
        Mark the run as started.
        """
        self.started_at = datetime.datetime.now().isoformat()
        self.status = "running"
        self._log_event("Run started")
    
    def complete(self) -> None:
        """
        Mark the run as completed.
        """
        self.ended_at = datetime.datetime.now().isoformat()
        self.status = "completed"
        self._log_event("Run completed")
    
    def fail(self, error_message: str) -> None:
        """
        Mark the run as failed.
        
        Args:
            error_message: Error message describing the failure.
        """
        self.ended_at = datetime.datetime.now().isoformat()
        self.status = "failed"
        self._log_event(f"Run failed: {error_message}")
        self.add_metadata("error_message", error_message)
    
    def log_parameter(self, key: str, value: Any) -> None:
        """
        Log a parameter for this run.
        
        Args:
            key: Parameter name.
            value: Parameter value.
        """
        self.parameters[key] = value
        self._log_event(f"Parameter logged: {key}={value}")
    
    def log_parameters(self, params: Dict[str, Any]) -> None:
        """
        Log multiple parameters for this run.
        
        Args:
            params: Dictionary of parameter names and values.
        """
        for key, value in params.items():
            self.log_parameter(key, value)
    
    def log_metric(self, key: str, value: Union[int, float]) -> None:
        """
        Log a metric for this run.
        
        Args:
            key: Metric name.
            value: Metric value.
            
        Raises:
            TrackingError: If the value is not a number.
        """
        if not isinstance(value, (int, float)):
            raise TrackingError(f"Metric value must be a number, got {type(value)}")
        
        self.metrics[key] = value
        self._log_event(f"Metric logged: {key}={value}")
    
    def log_metrics(self, metrics: Dict[str, Union[int, float]]) -> None:
        """
        Log multiple metrics for this run.
        
        Args:
            metrics: Dictionary of metric names and values.
        """
        for key, value in metrics.items():
            self.log_metric(key, value)
    
    def log_metric_step(self, key: str, value: Union[int, float], step: int) -> None:
        """
        Log a metric with a step for this run.
        
        Args:
            key: Metric name.
            value: Metric value.
            step: Step number.
            
        Raises:
            TrackingError: If the value is not a number.
        """
        if not isinstance(value, (int, float)):
            raise TrackingError(f"Metric value must be a number, got {type(value)}")
        
        if key not in self.history:
            self.history[key] = []
        
        self.history[key].append({"step": step, "value": value})
        self.metrics[key] = value  # Update the latest value
        self._log_event(f"Metric logged: {key}={value} (step {step})")
    
    def log_artifact(self, key: str, artifact: Any, artifact_path: Optional[str] = None) -> str:
        """
        Log an artifact for this run.
        
        Args:
            key: Artifact name.
            artifact: Artifact object.
            artifact_path: Optional path to save the artifact. If None, the artifact is stored in memory.
            
        Returns:
            Path to the saved artifact if artifact_path is provided, otherwise "memory".
        """
        if artifact_path:
            # Save the artifact to a file
            os.makedirs(os.path.dirname(artifact_path), exist_ok=True)
            
            if isinstance(artifact, plt.Figure):
                # Save matplotlib figure
                artifact.savefig(artifact_path)
                location = artifact_path
            elif isinstance(artifact, (np.ndarray, list, dict, str, int, float, bool)):
                # Save simple objects as JSON if possible
                try:
                    with open(artifact_path, "w") as f:
                        json.dump(artifact, f)
                    location = artifact_path
                except (TypeError, OverflowError):
                    # Fall back to pickle for objects that can't be JSON serialized
                    with open(artifact_path, "wb") as f:
                        pickle.dump(artifact, f)
                    location = artifact_path
            else:
                # Use pickle for other objects
                with open(artifact_path, "wb") as f:
                    pickle.dump(artifact, f)
                location = artifact_path
            
            self.artifacts[key] = {"type": type(artifact).__name__, "location": location}
        else:
            # Store the artifact in memory
            self.artifacts[key] = {"type": type(artifact).__name__, "location": "memory", "value": artifact}
            location = "memory"
        
        self._log_event(f"Artifact logged: {key} ({type(artifact).__name__})")
        return location
    
    def get_artifact(self, key: str) -> Any:
        """
        Get an artifact by name.
        
        Args:
            key: Artifact name.
            
        Returns:
            Artifact object if found.
            
        Raises:
            TrackingError: If the artifact is not found or cannot be loaded.
        """
        if key not in self.artifacts:
            raise TrackingError(f"Artifact not found: {key}")
        
        artifact_info = self.artifacts[key]
        
        if artifact_info["location"] == "memory":
            return artifact_info["value"]
        else:
            try:
                location = artifact_info["location"]
                
                if location.endswith((".json", ".txt")):
                    with open(location, "r") as f:
                        return json.load(f)
                else:
                    with open(location, "rb") as f:
                        return pickle.load(f)
            except Exception as e:
                raise TrackingError(f"Error loading artifact {key}: {str(e)}")
    
    def log_model(self, model: Any, model_path: str) -> str:
        """
        Log a model for this run.
        
        Args:
            model: Model object.
            model_path: Path to save the model.
            
        Returns:
            Path to the saved model.
        """
        os.makedirs(os.path.dirname(model_path), exist_ok=True)
        
        with open(model_path, "wb") as f:
            pickle.dump(model, f)
        
        self.artifacts["model"] = {"type": type(model).__name__, "location": model_path}
        self._log_event(f"Model logged: {type(model).__name__}")
        return model_path
    
    def get_model(self) -> Any:
        """
        Get the model for this run.
        
        Returns:
            Model object if found.
            
        Raises:
            TrackingError: If the model is not found or cannot be loaded.
        """
        return self.get_artifact("model")
    
    def log_message(self, message: str) -> None:
        """
        Log a message for this run.
        
        Args:
            message: Message to log.
        """
        self._log_event(message)
    
    def _log_event(self, message: str) -> None:
        """
        Log an event for this run.
        
        Args:
            message: Event message.
        """
        timestamp = datetime.datetime.now().isoformat()
        self.logs.append({"timestamp": timestamp, "message": message})
    
    def add_metadata(self, key: str, value: Any) -> None:
        """
        Add metadata to the run.
        
        Args:
            key: Metadata key.
            value: Metadata value.
        """
        self.metadata[key] = value
    
    def get_metadata(self, key: str) -> Any:
        """
        Get metadata from the run.
        
        Args:
            key: Metadata key.
            
        Returns:
            Metadata value if found, None otherwise.
        """
        return self.metadata.get(key)
    
    def to_dict(self) -> Dict[str, Any]:
        """
        Convert the run to a dictionary.
        
        Returns:
            Dictionary representation of the run.
        """
        # Filter out in-memory artifacts for serialization
        serializable_artifacts = {}
        for key, artifact_info in self.artifacts.items():
            if artifact_info["location"] == "memory":
                serializable_artifacts[key] = {
                    "type": artifact_info["type"],
                    "location": "memory"
                }
            else:
                serializable_artifacts[key] = artifact_info
        
        return {
            "run_id": self.run_id,
            "experiment_id": self.experiment_id,
            "name": self.name,
            "description": self.description,
            "tags": self.tags,
            "created_at": self.created_at,
            "started_at": self.started_at,
            "ended_at": self.ended_at,
            "status": self.status,
            "parameters": self.parameters,
            "metrics": self.metrics,
            "artifacts": serializable_artifacts,
            "logs": self.logs,
            "metadata": self.metadata,
            "history": self.history
        }
    
    @classmethod
    def from_dict(cls, data: Dict[str, Any]):
        """
        Create a run from a dictionary.
        
        Args:
            data: Dictionary representation of the run.
            
        Returns:
            Run object.
        """
        run = cls(
            name=data["name"],
            description=data.get("description", ""),
            tags=data.get("tags", [])
        )
        
        run.run_id = data["run_id"]
        run.experiment_id = data.get("experiment_id")
        run.created_at = data["created_at"]
        run.started_at = data.get("started_at")
        run.ended_at = data.get("ended_at")
        run.status = data.get("status", "created")
        run.parameters = data.get("parameters", {})
        run.metrics = data.get("metrics", {})
        run.artifacts = data.get("artifacts", {})
        run.logs = data.get("logs", [])
        run.metadata = data.get("metadata", {})
        run.history = data.get("history", {})
        
        return run
    
    def save(self, directory: str) -> str:
        """
        Save the run to a file.
        
        Args:
            directory: Directory to save the run to.
            
        Returns:
            Path to the saved run file.
        """
        os.makedirs(directory, exist_ok=True)
        
        file_path = os.path.join(directory, f"{self.run_id}.json")
        
        with open(file_path, "w") as f:
            json.dump(self.to_dict(), f, indent=2)
        
        return file_path
    
    @classmethod
    def load(cls, file_path: str):
        """
        Load a run from a file.
        
        Args:
            file_path: Path to the run file.
            
        Returns:
            Run object.
            
        Raises:
            FileNotFoundError: If the file does not exist.
            json.JSONDecodeError: If the file is not valid JSON.
        """
        with open(file_path, "r") as f:
            data = json.load(f)
        
        return cls.from_dict(data)
    
    def __repr__(self) -> str:
        """
        Get string representation of the run.
        
        Returns:
            String representation.
        """
        return f"Run(id={self.run_id}, name={self.name}, status={self.status})"



"""
Experiment class for SBYB tracking.

This module provides the Experiment class, which represents a collection of related runs
for a machine learning experiment.
"""

from typing import Any, Dict, List, Optional, Union
import os
import json
import datetime
import uuid
from pathlib import Path

from sbyb.core.base import SBYBComponent
from sbyb.core.exceptions import TrackingError


class Experiment(SBYBComponent):
    """
    Represents a collection of related runs for a machine learning experiment.
    
    An experiment groups together multiple runs that are part of the same
    machine learning task or investigation. It provides methods for managing
    runs, tracking metadata, and comparing results.
    """
    
    def __init__(self, name: str, description: Optional[str] = None, 
                tags: Optional[List[str]] = None, config: Optional[Dict[str, Any]] = None):
        """
        Initialize a new experiment.
        
        Args:
            name: Name of the experiment.
            description: Optional description of the experiment.
            tags: Optional list of tags for categorizing the experiment.
            config: Optional configuration dictionary.
        """
        super().__init__(config)
        
        self.name = name
        self.description = description or ""
        self.tags = tags or []
        self.experiment_id = str(uuid.uuid4())
        self.created_at = datetime.datetime.now().isoformat()
        self.runs = []
        self.metadata = {}
        
    def add_run(self, run) -> None:
        """
        Add a run to this experiment.
        
        Args:
            run: Run object to add to this experiment.
        """
        run.experiment_id = self.experiment_id
        self.runs.append(run)
    
    def get_run(self, run_id: str):
        """
        Get a run by ID.
        
        Args:
            run_id: ID of the run to retrieve.
            
        Returns:
            Run object if found, None otherwise.
        """
        for run in self.runs:
            if run.run_id == run_id:
                return run
        return None
    
    def get_runs(self) -> List:
        """
        Get all runs in this experiment.
        
        Returns:
            List of Run objects.
        """
        return self.runs
    
    def get_best_run(self, metric: str, higher_is_better: bool = True):
        """
        Get the best run based on a metric.
        
        Args:
            metric: Name of the metric to use for comparison.
            higher_is_better: Whether higher values of the metric are better.
            
        Returns:
            Best Run object if found, None otherwise.
            
        Raises:
            TrackingError: If no runs have the specified metric.
        """
        if not self.runs:
            return None
        
        valid_runs = [run for run in self.runs if metric in run.metrics]
        
        if not valid_runs:
            raise TrackingError(f"No runs have the metric '{metric}'")
        
        if higher_is_better:
            return max(valid_runs, key=lambda run: run.metrics[metric])
        else:
            return min(valid_runs, key=lambda run: run.metrics[metric])
    
    def add_metadata(self, key: str, value: Any) -> None:
        """
        Add metadata to the experiment.
        
        Args:
            key: Metadata key.
            value: Metadata value.
        """
        self.metadata[key] = value
    
    def get_metadata(self, key: str) -> Any:
        """
        Get metadata from the experiment.
        
        Args:
            key: Metadata key.
            
        Returns:
            Metadata value if found, None otherwise.
        """
        return self.metadata.get(key)
    
    def to_dict(self) -> Dict[str, Any]:
        """
        Convert the experiment to a dictionary.
        
        Returns:
            Dictionary representation of the experiment.
        """
        return {
            "experiment_id": self.experiment_id,
            "name": self.name,
            "description": self.description,
            "tags": self.tags,
            "created_at": self.created_at,
            "metadata": self.metadata,
            "runs": [run.to_dict() for run in self.runs]
        }
    
    @classmethod
    def from_dict(cls, data: Dict[str, Any]):
        """
        Create an experiment from a dictionary.
        
        Args:
            data: Dictionary representation of the experiment.
            
        Returns:
            Experiment object.
        """
        from sbyb.tracking.run import Run
        
        experiment = cls(
            name=data["name"],
            description=data.get("description", ""),
            tags=data.get("tags", [])
        )
        
        experiment.experiment_id = data["experiment_id"]
        experiment.created_at = data["created_at"]
        experiment.metadata = data.get("metadata", {})
        
        for run_data in data.get("runs", []):
            run = Run.from_dict(run_data)
            experiment.runs.append(run)
        
        return experiment
    
    def save(self, directory: str) -> str:
        """
        Save the experiment to a file.
        
        Args:
            directory: Directory to save the experiment to.
            
        Returns:
            Path to the saved experiment file.
        """
        os.makedirs(directory, exist_ok=True)
        
        file_path = os.path.join(directory, f"{self.experiment_id}.json")
        
        with open(file_path, "w") as f:
            json.dump(self.to_dict(), f, indent=2)
        
        return file_path
    
    @classmethod
    def load(cls, file_path: str):
        """
        Load an experiment from a file.
        
        Args:
            file_path: Path to the experiment file.
            
        Returns:
            Experiment object.
            
        Raises:
            FileNotFoundError: If the file does not exist.
            json.JSONDecodeError: If the file is not valid JSON.
        """
        with open(file_path, "r") as f:
            data = json.load(f)
        
        return cls.from_dict(data)
    
    def compare_runs(self, metrics: List[str]) -> Dict[str, Dict[str, Any]]:
        """
        Compare runs based on specified metrics.
        
        Args:
            metrics: List of metric names to compare.
            
        Returns:
            Dictionary mapping run IDs to dictionaries of metric values.
        """
        result = {}
        
        for run in self.runs:
            run_metrics = {}
            for metric in metrics:
                if metric in run.metrics:
                    run_metrics[metric] = run.metrics[metric]
                else:
                    run_metrics[metric] = None
            
            result[run.run_id] = {
                "metrics": run_metrics,
                "parameters": run.parameters,
                "name": run.name,
                "status": run.status
            }
        
        return result
    
    def __repr__(self) -> str:
        """
        Get string representation of the experiment.
        
        Returns:
            String representation.
        """
        return f"Experiment(id={self.experiment_id}, name={self.name}, runs={len(self.runs)})"



"""
Main task detection component for SBYB.

This module provides the main task detection logic for automatically determining
the appropriate machine learning task based on data characteristics.
"""

from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import os
from pathlib import Path

from sbyb.core.base import TaskDetector, SBYBComponent
from sbyb.core.exceptions import TaskDetectionError
from sbyb.core.utils import get_column_types, is_classification_target, is_regression_target


class TaskDetector(SBYBComponent):
    """
    Task detection component.
    
    This component automatically detects the appropriate machine learning task
    based on the data characteristics.
    """
    
    TASK_TYPES = [
        'binary_classification',
        'multiclass_classification',
        'multilabel_classification',
        'regression',
        'clustering',
        'nlp_classification',
        'nlp_regression',
        'computer_vision',
        'time_series_forecasting',
        'time_series_classification',
        'anomaly_detection'
    ]
    
    def __init__(self, threshold: float = 0.7, default_task: str = 'classification', config: Optional[Dict[str, Any]] = None):
        """
        Initialize the task detector.
        
        Args:
            threshold: Confidence threshold for task detection.
            default_task: Default task to use if detection is uncertain.
            config: Configuration dictionary for the task detector.
        """
        super().__init__(config)
        self.threshold = threshold
        self.default_task = default_task
    
    def detect(self, data: Union[pd.DataFrame, np.ndarray, str], target: Optional[Union[str, int]] = None) -> Dict[str, Any]:
        """
        Detect the machine learning task type from the data.
        
        Args:
            data: Input data or path to data.
            target: Target variable name or index.
            
        Returns:
            Dictionary with detected task information:
                - task_type: The detected task type
                - confidence: Confidence score for the detection
                - data_type: The detected data type (tabular, text, image, time_series)
                - details: Additional details about the detection
        """
        # Handle different input types
        data_type, processed_data = self._process_input(data)
        
        # Detect task based on data type
        if data_type == 'tabular':
            return self._detect_tabular_task(processed_data, target)
        elif data_type == 'text':
            return self._detect_text_task(processed_data, target)
        elif data_type == 'image':
            return self._detect_image_task(processed_data, target)
        elif data_type == 'time_series':
            return self._detect_time_series_task(processed_data, target)
        else:
            raise TaskDetectionError(f"Unsupported data type: {data_type}")
    
    def _process_input(self, data: Union[pd.DataFrame, np.ndarray, str]) -> Tuple[str, Any]:
        """
        Process the input data and determine its type.
        
        Args:
            data: Input data or path to data.
            
        Returns:
            Tuple of (data_type, processed_data)
        """
        # If data is a string, it's a path to a file or directory
        if isinstance(data, str):
            return self._detect_from_path(data)
        
        # If data is a DataFrame, it's tabular data
        if isinstance(data, pd.DataFrame):
            return 'tabular', data
        
        # If data is a numpy array, determine its type based on shape
        if isinstance(data, np.ndarray):
            if len(data.shape) == 2:
                # 2D array is likely tabular data
                return 'tabular', pd.DataFrame(data)
            elif len(data.shape) == 3:
                # 3D array could be time series or grayscale images
                if data.shape[2] == 1:
                    return 'image', data
                else:
                    return 'time_series', data
            elif len(data.shape) == 4:
                # 4D array is likely image data (batch, height, width, channels)
                return 'image', data
        
        raise TaskDetectionError(f"Unable to determine data type from input: {type(data)}")
    
    def _detect_from_path(self, path: str) -> Tuple[str, Any]:
        """
        Detect data type from a file or directory path.
        
        Args:
            path: Path to file or directory.
            
        Returns:
            Tuple of (data_type, processed_data)
        """
        if os.path.isdir(path):
            # Check if it's an image directory
            image_extensions = ['.jpg', '.jpeg', '.png', '.bmp', '.gif', '.tiff']
            files = list(Path(path).glob('*'))
            image_files = [f for f in files if f.suffix.lower() in image_extensions]
            
            if len(image_files) > 0 and len(image_files) / len(files) > 0.5:
                return 'image', path
            
            # Check if it's a text directory
            text_extensions = ['.txt', '.csv', '.json', '.xml', '.html']
            text_files = [f for f in files if f.suffix.lower() in text_extensions]
            
            if len(text_files) > 0 and len(text_files) / len(files) > 0.5:
                return 'text', path
        else:
            # It's a file, determine type from extension
            file_extension = os.path.splitext(path)[1].lower()
            
            if file_extension in ['.csv', '.xlsx', '.xls']:
                # Load tabular data
                if file_extension == '.csv':
                    data = pd.read_csv(path)
                else:
                    data = pd.read_excel(path)
                return 'tabular', data
            
            elif file_extension in ['.txt', '.json', '.xml', '.html']:
                return 'text', path
            
            elif file_extension in ['.jpg', '.jpeg', '.png', '.bmp', '.gif', '.tiff']:
                return 'image', path
        
        # Default to tabular and try to load as CSV
        try:
            data = pd.read_csv(path)
            return 'tabular', data
        except:
            raise TaskDetectionError(f"Unable to determine data type from path: {path}")
    
    def _detect_tabular_task(self, data: pd.DataFrame, target: Optional[Union[str, int]] = None) -> Dict[str, Any]:
        """
        Detect the task type for tabular data.
        
        Args:
            data: Input tabular data.
            target: Target variable name or index.
            
        Returns:
            Dictionary with detected task information.
        """
        # If no target is specified, assume clustering or try to guess the target
        if target is None:
            # Try to find a likely target column (e.g., 'target', 'label', 'y', etc.)
            likely_targets = ['target', 'label', 'class', 'y', 'output', 'result']
            for col in likely_targets:
                if col in data.columns:
                    target = col
                    break
            
            # If still no target, check if there's a column that looks like a target
            if target is None:
                # Look for columns with few unique values relative to the dataset size
                for col in data.columns:
                    if pd.api.types.is_numeric_dtype(data[col].dtype):
                        n_unique = data[col].nunique()
                        if 1 < n_unique <= min(20, len(data) * 0.1):
                            target = col
                            break
            
            # If still no target, assume clustering
            if target is None:
                return {
                    'task_type': 'clustering',
                    'confidence': 0.8,
                    'data_type': 'tabular',
                    'details': {
                        'reason': 'No target variable specified or detected'
                    }
                }
        
        # Extract target variable
        if isinstance(target, int):
            y = data.iloc[:, target]
            target_name = data.columns[target]
        else:
            y = data[target]
            target_name = target
        
        # Check if it's a classification task
        if is_classification_target(y):
            n_classes = y.nunique()
            
            if n_classes == 2:
                return {
                    'task_type': 'binary_classification',
                    'confidence': 0.9,
                    'data_type': 'tabular',
                    'details': {
                        'n_classes': n_classes,
                        'target_column': target_name
                    }
                }
            else:
                return {
                    'task_type': 'multiclass_classification',
                    'confidence': 0.9,
                    'data_type': 'tabular',
                    'details': {
                        'n_classes': n_classes,
                        'target_column': target_name
                    }
                }
        
        # Check if it's a regression task
        if is_regression_target(y):
            return {
                'task_type': 'regression',
                'confidence': 0.9,
                'data_type': 'tabular',
                'details': {
                    'target_column': target_name
                }
            }
        
        # If we can't determine the task, default to classification
        return {
            'task_type': self.default_task,
            'confidence': 0.5,
            'data_type': 'tabular',
            'details': {
                'reason': 'Unable to confidently determine task type',
                'target_column': target_name
            }
        }
    
    def _detect_text_task(self, data: Union[pd.DataFrame, str], target: Optional[Union[str, int]] = None) -> Dict[str, Any]:
        """
        Detect the task type for text data.
        
        Args:
            data: Input text data or path to text data.
            target: Target variable name or index.
            
        Returns:
            Dictionary with detected task information.
        """
        # If data is a DataFrame with a target column, determine if it's classification or regression
        if isinstance(data, pd.DataFrame) and target is not None:
            # Extract target variable
            if isinstance(target, int):
                y = data.iloc[:, target]
                target_name = data.columns[target]
            else:
                y = data[target]
                target_name = target
            
            # Check if it's a classification task
            if is_classification_target(y):
                n_classes = y.nunique()
                
                if n_classes == 2:
                    return {
                        'task_type': 'nlp_classification',
                        'confidence': 0.9,
                        'data_type': 'text',
                        'details': {
                            'n_classes': n_classes,
                            'target_column': target_name,
                            'subtask': 'binary_classification'
                        }
                    }
                else:
                    return {
                        'task_type': 'nlp_classification',
                        'confidence': 0.9,
                        'data_type': 'text',
                        'details': {
                            'n_classes': n_classes,
                            'target_column': target_name,
                            'subtask': 'multiclass_classification'
                        }
                    }
            
            # Check if it's a regression task
            if is_regression_target(y):
                return {
                    'task_type': 'nlp_regression',
                    'confidence': 0.9,
                    'data_type': 'text',
                    'details': {
                        'target_column': target_name
                    }
                }
        
        # If we can't determine the task, default to NLP classification
        return {
            'task_type': 'nlp_classification',
            'confidence': 0.7,
            'data_type': 'text',
            'details': {
                'reason': 'Default NLP task without specific target information'
            }
        }
    
    def _detect_image_task(self, data: Union[np.ndarray, str], target: Optional[Union[str, int]] = None) -> Dict[str, Any]:
        """
        Detect the task type for image data.
        
        Args:
            data: Input image data or path to image data.
            target: Target variable name or index.
            
        Returns:
            Dictionary with detected task information.
        """
        # For image data, we typically assume computer vision tasks
        # If we have target information, we can determine if it's classification or regression
        if target is not None and isinstance(data, str) and os.path.isdir(data):
            # Check if subdirectories exist (common for image classification datasets)
            subdirs = [d for d in os.listdir(data) if os.path.isdir(os.path.join(data, d))]
            
            if len(subdirs) > 0:
                return {
                    'task_type': 'computer_vision',
                    'confidence': 0.9,
                    'data_type': 'image',
                    'details': {
                        'subtask': 'classification',
                        'n_classes': len(subdirs),
                        'classes': subdirs
                    }
                }
        
        # Default to general computer vision
        return {
            'task_type': 'computer_vision',
            'confidence': 0.8,
            'data_type': 'image',
            'details': {
                'reason': 'Default computer vision task without specific target information'
            }
        }
    
    def _detect_time_series_task(self, data: Union[pd.DataFrame, np.ndarray], target: Optional[Union[str, int]] = None) -> Dict[str, Any]:
        """
        Detect the task type for time series data.
        
        Args:
            data: Input time series data.
            target: Target variable name or index.
            
        Returns:
            Dictionary with detected task information.
        """
        # If data is a DataFrame, check for datetime columns
        if isinstance(data, pd.DataFrame):
            # Check for datetime columns
            datetime_cols = [col for col in data.columns if pd.api.types.is_datetime64_any_dtype(data[col].dtype)]
            
            if len(datetime_cols) > 0:
                # If we have a target column, determine if it's forecasting or classification
                if target is not None:
                    # Extract target variable
                    if isinstance(target, int):
                        y = data.iloc[:, target]
                        target_name = data.columns[target]
                    else:
                        y = data[target]
                        target_name = target
                    
                    # Check if it's a classification task
                    if is_classification_target(y):
                        return {
                            'task_type': 'time_series_classification',
                            'confidence': 0.9,
                            'data_type': 'time_series',
                            'details': {
                                'datetime_columns': datetime_cols,
                                'target_column': target_name,
                                'n_classes': y.nunique()
                            }
                        }
                    
                    # Otherwise, assume forecasting
                    return {
                        'task_type': 'time_series_forecasting',
                        'confidence': 0.9,
                        'data_type': 'time_series',
                        'details': {
                            'datetime_columns': datetime_cols,
                            'target_column': target_name
                        }
                    }
        
        # Default to time series forecasting
        return {
            'task_type': 'time_series_forecasting',
            'confidence': 0.7,
            'data_type': 'time_series',
            'details': {
                'reason': 'Default time series task without specific target information'
            }
        }



"""
Feature scaling for SBYB.

This module provides components for scaling numeric features.
"""

from typing import Any, Dict, List, Optional, Union

import numpy as np
import pandas as pd
from sklearn.preprocessing import StandardScaler, MinMaxScaler, RobustScaler, MaxAbsScaler

from sbyb.core.exceptions import PreprocessingError
from sbyb.core.utils import get_column_types
from sbyb.preprocessing.base import ColumnBasedPreprocessor


class FeatureScaler(ColumnBasedPreprocessor):
    """
    Feature scaling component.
    
    This component scales numeric features using various methods.
    """
    
    SCALING_METHODS = ['standard', 'minmax', 'robust', 'maxabs']
    
    def __init__(self, columns: Optional[List[str]] = None,
                 method: str = 'standard',
                 with_mean: bool = True,
                 with_std: bool = True,
                 feature_range: tuple = (0, 1),
                 config: Optional[Dict[str, Any]] = None):
        """
        Initialize the feature scaler.
        
        Args:
            columns: List of column names to scale. If None, all numeric columns will be processed.
            method: Scaling method ('standard', 'minmax', 'robust', 'maxabs').
            with_mean: Whether to center the data before scaling (for standard scaling).
            with_std: Whether to scale the data to unit variance (for standard scaling).
            feature_range: Range of transformed values (for minmax scaling).
            config: Configuration dictionary for the scaler.
        """
        super().__init__(columns, config)
        
        # Validate parameters
        if method not in self.SCALING_METHODS:
            raise ValueError(f"Invalid scaling method: {method}. Must be one of {self.SCALING_METHODS}")
        
        self.method = method
        self.with_mean = with_mean
        self.with_std = with_std
        self.feature_range = feature_range
        
        # Scaler for all columns
        self._scaler = None
    
    def _select_columns(self, data: pd.DataFrame) -> List[str]:
        """
        Select numeric columns for scaling.
        
        Args:
            data: Input data.
            
        Returns:
            List of numeric column names.
        """
        column_types = get_column_types(data)
        return column_types['numeric']
    
    def fit(self, data: Union[pd.DataFrame, np.ndarray], **kwargs) -> 'FeatureScaler':
        """
        Fit the feature scaler to the data.
        
        Args:
            data: Input data to fit the scaler.
            **kwargs: Additional keyword arguments.
            
        Returns:
            self: The fitted scaler.
        """
        data = self._validate_input(data)
        self._fitted_columns = self._get_columns(data)
        
        # Create scaler based on method
        if self.method == 'standard':
            self._scaler = StandardScaler(with_mean=self.with_mean, with_std=self.with_std)
        elif self.method == 'minmax':
            self._scaler = MinMaxScaler(feature_range=self.feature_range)
        elif self.method == 'robust':
            self._scaler = RobustScaler(with_centering=self.with_mean, with_scaling=self.with_std)
        elif self.method == 'maxabs':
            self._scaler = MaxAbsScaler()
        
        # Fit scaler on selected columns
        if self._fitted_columns:
            self._scaler.fit(data[self._fitted_columns])
        
        self._is_fitted = True
        return self
    
    def transform(self, data: Union[pd.DataFrame, np.ndarray], **kwargs) -> pd.DataFrame:
        """
        Scale numeric features in the data.
        
        Args:
            data: Input data to scale.
            **kwargs: Additional keyword arguments.
            
        Returns:
            Data with scaled numeric features.
        """
        self._check_is_fitted()
        data = self._validate_input(data).copy()
        
        # Scale selected columns
        if self._fitted_columns and self._scaler:
            columns_to_scale = [col for col in self._fitted_columns if col in data.columns]
            if columns_to_scale:
                scaled_data = self._scaler.transform(data[columns_to_scale])
                for i, col in enumerate(columns_to_scale):
                    data[col] = scaled_data[:, i]
        
        return data



"""
NLP task detection for SBYB.

This module provides specialized components for detecting NLP tasks.
"""

from typing import Any, Dict, List, Optional, Union

import numpy as np
import pandas as pd
import os
import re

from sbyb.core.base import SBYBComponent
from sbyb.core.exceptions import TaskDetectionError


class NLPDetector(SBYBComponent):
    """
    NLP task detection component.
    
    This component specializes in detecting natural language processing tasks.
    """
    
    NLP_TASKS = [
        'text_classification',
        'sentiment_analysis',
        'named_entity_recognition',
        'text_generation',
        'summarization',
        'translation',
        'question_answering',
        'text_regression'
    ]
    
    def __init__(self, config: Optional[Dict[str, Any]] = None):
        """
        Initialize the NLP detector.
        
        Args:
            config: Configuration dictionary for the detector.
        """
        super().__init__(config)
    
    def detect(self, data: Union[pd.DataFrame, str], text_column: Optional[str] = None, 
               target: Optional[Union[str, int]] = None) -> Dict[str, Any]:
        """
        Detect if the data represents an NLP task and identify the specific task type.
        
        Args:
            data: Input data or path to text data.
            text_column: Column name containing text data (if data is a DataFrame).
            target: Target variable name or index (if applicable).
            
        Returns:
            Dictionary with detection results:
                - is_nlp: Whether the task is an NLP task
                - task_type: Specific NLP task type
                - confidence: Confidence score for the detection
                - details: Additional details about the detection
        """
        # Handle different input types
        if isinstance(data, str):
            # Path to a file or directory
            return self._detect_from_path(data, target)
        elif isinstance(data, pd.DataFrame):
            # DataFrame with text data
            return self._detect_from_dataframe(data, text_column, target)
        else:
            raise TaskDetectionError(f"Unsupported data type for NLP detection: {type(data)}")
    
    def _detect_from_path(self, path: str, target: Optional[Union[str, int]] = None) -> Dict[str, Any]:
        """
        Detect NLP task from a file or directory path.
        
        Args:
            path: Path to file or directory.
            target: Target variable name or index (if applicable).
            
        Returns:
            Dictionary with detection results.
        """
        if os.path.isdir(path):
            # Directory with text files
            # Check for common NLP dataset structures
            subdirs = [d for d in os.listdir(path) if os.path.isdir(os.path.join(path, d))]
            
            if len(subdirs) > 0:
                # Likely a text classification dataset with subdirectories for classes
                return {
                    'is_nlp': True,
                    'task_type': 'text_classification',
                    'confidence': 0.8,
                    'details': {
                        'n_classes': len(subdirs),
                        'classes': subdirs,
                        'data_source': 'directory'
                    }
                }
            else:
                # Generic text data
                return {
                    'is_nlp': True,
                    'task_type': 'text_classification',  # Default assumption
                    'confidence': 0.6,
                    'details': {
                        'data_source': 'directory',
                        'note': 'No clear class structure, defaulting to text classification'
                    }
                }
        else:
            # Single file
            file_extension = os.path.splitext(path)[1].lower()
            
            if file_extension in ['.txt', '.csv', '.json', '.xml']:
                # Try to read a sample of the file to determine content
                try:
                    with open(path, 'r', encoding='utf-8') as f:
                        sample = f.read(10000)  # Read first 10KB
                    
                    # Check for common patterns
                    has_long_text = any(len(line) > 100 for line in sample.split('\n') if line.strip())
                    has_qa_pattern = '?' in sample and len(re.findall(r'\?\s+[A-Z]', sample)) > 0
                    has_sentiment_words = any(word in sample.lower() for word in ['positive', 'negative', 'sentiment', 'rating', 'stars'])
                    
                    if has_qa_pattern:
                        task_type = 'question_answering'
                        confidence = 0.7
                    elif has_sentiment_words:
                        task_type = 'sentiment_analysis'
                        confidence = 0.7
                    elif has_long_text:
                        task_type = 'summarization'
                        confidence = 0.6
                    else:
                        task_type = 'text_classification'
                        confidence = 0.5
                    
                    return {
                        'is_nlp': True,
                        'task_type': task_type,
                        'confidence': confidence,
                        'details': {
                            'data_source': 'file',
                            'file_type': file_extension,
                            'has_long_text': has_long_text,
                            'has_qa_pattern': has_qa_pattern,
                            'has_sentiment_words': has_sentiment_words
                        }
                    }
                except:
                    # If reading fails, return a generic result
                    return {
                        'is_nlp': True,
                        'task_type': 'text_classification',  # Default assumption
                        'confidence': 0.5,
                        'details': {
                            'data_source': 'file',
                            'file_type': file_extension,
                            'note': 'Could not analyze file content'
                        }
                    }
            else:
                return {
                    'is_nlp': False,
                    'confidence': 0.7,
                    'details': {
                        'reason': f'Unsupported file type for NLP: {file_extension}',
                        'data_source': 'file'
                    }
                }
    
    def _detect_from_dataframe(self, data: pd.DataFrame, text_column: Optional[str] = None, 
                              target: Optional[Union[str, int]] = None) -> Dict[str, Any]:
        """
        Detect NLP task from a DataFrame.
        
        Args:
            data: Input DataFrame.
            text_column: Column name containing text data.
            target: Target variable name or index (if applicable).
            
        Returns:
            Dictionary with detection results.
        """
        # If text_column is not specified, try to find it
        if text_column is None:
            text_column = self._find_text_column(data)
            
            if text_column is None:
                return {
                    'is_nlp': False,
                    'confidence': 0.8,
                    'details': {
                        'reason': 'No suitable text column found in the DataFrame'
                    }
                }
        
        # Verify that the text column exists
        if text_column not in data.columns:
            return {
                'is_nlp': False,
                'confidence': 0.9,
                'details': {
                    'reason': f'Specified text column "{text_column}" not found in the DataFrame'
                }
            }
        
        # Analyze text characteristics
        text_data = data[text_column].astype(str)
        avg_length = text_data.str.len().mean()
        max_length = text_data.str.len().max()
        has_long_text = avg_length > 100 or max_length > 500
        
        # If target is specified, determine the task type
        if target is not None:
            # Extract target variable
            if isinstance(target, int):
                y = data.iloc[:, target]
                target_name = data.columns[target]
            else:
                y = data[target]
                target_name = target
            
            # Check if it's classification or regression
            is_categorical = pd.api.types.is_categorical_dtype(y.dtype) or pd.api.types.is_string_dtype(y.dtype)
            n_unique = y.nunique()
            is_classification = is_categorical or (pd.api.types.is_numeric_dtype(y.dtype) and n_unique <= min(10, len(y) * 0.05))
            
            if is_classification:
                # Check for sentiment analysis patterns
                has_sentiment_pattern = any(col.lower() in ['sentiment', 'rating', 'stars', 'score', 'emotion'] 
                                           for col in data.columns)
                
                if has_sentiment_pattern:
                    task_type = 'sentiment_analysis'
                    confidence = 0.8
                else:
                    task_type = 'text_classification'
                    confidence = 0.9
                
                return {
                    'is_nlp': True,
                    'task_type': task_type,
                    'confidence': confidence,
                    'details': {
                        'text_column': text_column,
                        'target_column': target_name,
                        'n_classes': n_unique,
                        'avg_text_length': avg_length,
                        'has_long_text': has_long_text,
                        'has_sentiment_pattern': has_sentiment_pattern if 'has_sentiment_pattern' in locals() else False
                    }
                }
            else:
                # Regression task
                return {
                    'is_nlp': True,
                    'task_type': 'text_regression',
                    'confidence': 0.9,
                    'details': {
                        'text_column': text_column,
                        'target_column': target_name,
                        'avg_text_length': avg_length,
                        'has_long_text': has_long_text
                    }
                }
        
        # If no target is specified, try to guess the task type from text characteristics
        has_qa_pattern = text_data.str.contains('\?').mean() > 0.2
        has_sentiment_words = text_data.str.contains('positive|negative|good|bad|excellent|terrible', 
                                                    case=False).mean() > 0.1
        
        if has_qa_pattern:
            task_type = 'question_answering'
            confidence = 0.7
        elif has_sentiment_words:
            task_type = 'sentiment_analysis'
            confidence = 0.7
        elif has_long_text:
            task_type = 'summarization'
            confidence = 0.6
        else:
            task_type = 'text_classification'
            confidence = 0.5
        
        return {
            'is_nlp': True,
            'task_type': task_type,
            'confidence': confidence,
            'details': {
                'text_column': text_column,
                'avg_text_length': avg_length,
                'has_long_text': has_long_text,
                'has_qa_pattern': has_qa_pattern,
                'has_sentiment_words': has_sentiment_words
            }
        }
    
    def _find_text_column(self, data: pd.DataFrame) -> Optional[str]:
        """
        Find the most likely text column in a DataFrame.
        
        Args:
            data: Input DataFrame.
            
        Returns:
            Name of the most likely text column, or None if no suitable column is found.
        """
        # Look for columns with string data type
        string_columns = [col for col in data.columns if pd.api.types.is_string_dtype(data[col].dtype)]
        
        if not string_columns:
            return None
        
        # Calculate average string length for each column
        avg_lengths = {}
        for col in string_columns:
            avg_lengths[col] = data[col].astype(str).str.len().mean()
        
        # Look for columns with common text-related names
        text_related_names = ['text', 'content', 'description', 'summary', 'comment', 'review', 'message', 'document']
        for name in text_related_names:
            for col in string_columns:
                if name in col.lower():
                    # If the column name contains a text-related term and has reasonable length, return it
                    if avg_lengths[col] > 20:
                        return col
        
        # If no column with text-related name is found, return the column with the longest average string length
        if avg_lengths:
            best_col = max(avg_lengths.items(), key=lambda x: x[1])
            if best_col[1] > 20:  # Only return if average length is reasonable
                return best_col[0]
        
        return None
    
    def suggest_preprocessing(self, data: Union[pd.DataFrame, str], text_column: Optional[str] = None) -> Dict[str, Any]:
        """
        Suggest preprocessing steps for the text data.
        
        Args:
            data: Input data or path to text data.
            text_column: Column name containing text data (if data is a DataFrame).
            
        Returns:
            Dictionary with suggested preprocessing steps.
        """
        # Get text data
        if isinstance(data, str):
            # For simplicity, just return general recommendations for file paths
            return {
                'suggested_steps': [
                    'tokenization',
                    'lowercasing',
                    'stopword_removal',
                    'stemming_or_lemmatization',
                    'vectorization'
                ],
                'vectorization_methods': [
                    {'method': 'tfidf', 'suitability': 'high'},
                    {'method': 'count', 'suitability': 'medium'},
                    {'method': 'word_embeddings', 'suitability': 'high'}
                ]
            }
        
        elif isinstance(data, pd.DataFrame):
            # If text_column is not specified, try to find it
            if text_column is None:
                text_column = self._find_text_column(data)
                
                if text_column is None:
                    return {
                        'error': 'No suitable text column found in the DataFrame'
                    }
            
            # Verify that the text column exists
            if text_column not in data.columns:
                return {
                    'error': f'Specified text column "{text_column}" not found in the DataFrame'
                }
            
            # Analyze text characteristics
            text_data = data[text_column].astype(str)
            avg_length = text_data.str.len().mean()
            max_length = text_data.str.len().max()
            has_long_text = avg_length > 100 or max_length > 500
            
            # Check for uppercase text
            uppercase_ratio = text_data.str.count(r'[A-Z]').sum() / text_data.str.count(r'[a-zA-Z]').sum()
            needs_lowercasing = uppercase_ratio > 0.2
            
            # Check for punctuation
            punctuation_ratio = text_data.str.count(r'[^\w\s]').sum() / text_data.str.len().sum()
            has_punctuation = punctuation_ratio > 0.05
            
            # Check for numbers
            number_ratio = text_data.str.count(r'\d').sum() / text_data.str.len().sum()
            has_numbers = number_ratio > 0.05
            
            # Suggest preprocessing steps
            steps = ['tokenization']
            
            if needs_lowercasing:
                steps.append('lowercasing')
            
            if has_punctuation:
                steps.append('punctuation_removal')
            
            if has_numbers and number_ratio < 0.2:  # If not too many numbers
                steps.append('number_removal')
            
            steps.extend(['stopword_removal', 'stemming_or_lemmatization'])
            
            # Suggest vectorization methods
            vectorization_methods = []
            
            if has_long_text:
                vectorization_methods.append({'method': 'tfidf', 'suitability': 'high'})
                vectorization_methods.append({'method': 'word_embeddings', 'suitability': 'high'})
                vectorization_methods.append({'method': 'count', 'suitability': 'low'})
            else:
                vectorization_methods.append({'method': 'tfidf', 'suitability': 'high'})
                vectorization_methods.append({'method': 'count', 'suitability': 'medium'})
                vectorization_methods.append({'method': 'word_embeddings', 'suitability': 'medium'})
            
            return {
                'suggested_steps': steps,
                'vectorization_methods': vectorization_methods,
                'details': {
                    'text_column': text_column,
                    'avg_length': avg_length,
                    'max_length': max_length,
                    'has_long_text': has_long_text,
                    'uppercase_ratio': uppercase_ratio,
                    'punctuation_ratio': punctuation_ratio,
                    'number_ratio': number_ratio
                }
            }
        
        else:
            return {
                'error': f'Unsupported data type for NLP preprocessing suggestions: {type(data)}'
            }
